# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import argparse
import unittest
from unittest import mock

from callee import Contains

from mozphab import exceptions
from mozphab.commands import abandon


class TestCheckRevisionId(unittest.TestCase):
    """Test the revision ID validation functionality."""

    def test_single_ids(self):
        """Test parsing single revision IDs."""
        # Basic integer
        result = abandon.check_revision_id("123")
        self.assertEqual(result, 123)

        # D-prefixed
        result = abandon.check_revision_id("D456")
        self.assertEqual(result, 456)

    def test_urls(self):
        """Test parsing Phabricator URLs."""
        # HTTPS URL
        result = abandon.check_revision_id(
            "https://phabricator.services.mozilla.com/D12345"
        )
        self.assertEqual(result, 12345)

        # HTTP URL
        result = abandon.check_revision_id("http://phabricator.example.com/D678")
        self.assertEqual(result, 678)

    def test_invalid_format(self):
        """Test that invalid formats raise errors."""
        with self.assertRaises(argparse.ArgumentTypeError) as cm:
            abandon.check_revision_id("invalid")
        self.assertIn("Invalid Revision ID", str(cm.exception))

        with self.assertRaises(argparse.ArgumentTypeError) as cm:
            abandon.check_revision_id("D")
        self.assertIn("Invalid Revision ID", str(cm.exception))

        with self.assertRaises(argparse.ArgumentTypeError) as cm:
            abandon.check_revision_id("123-125")
        self.assertIn("Invalid Revision ID", str(cm.exception))


class TestAbandonRevisions(unittest.TestCase):
    """Test the main abandon functionality."""

    def setUp(self):
        """Set up test fixtures."""
        self.args = argparse.Namespace(yes=True)  # Skip confirmation for tests
        self.mock_revisions = [
            {
                "id": 123,
                "phid": "PHID-DREV-123",
                "fields": {
                    "title": "Test revision 123",
                    "status": {"value": "needs-review"},
                },
            },
            {
                "id": 124,
                "phid": "PHID-DREV-124",
                "fields": {
                    "title": "Test revision 124",
                    "status": {"value": "abandoned"},
                },
            },
            {
                "id": 125,
                "phid": "PHID-DREV-125",
                "fields": {
                    "title": "Test revision 125",
                    "status": {"value": "accepted"},
                },
            },
        ]

    @mock.patch("mozphab.commands.abandon.conduit")
    def test_abandon_revisions_success(self, mock_conduit):
        """Test successful abandonment of revisions."""
        # Setup mocks
        mock_conduit.check.return_value = True
        mock_conduit.get_revisions.return_value = [
            self.mock_revisions[0],
            self.mock_revisions[2],
        ]

        # Call function
        abandon.abandon_revisions([123, 125], self.args)

        # Verify API calls
        mock_conduit.check.assert_called_once()
        mock_conduit.get_revisions.assert_called_once_with(ids=[123, 125])

        # Verify abandonment calls
        expected_calls = [
            mock.call(
                rev_id="PHID-DREV-123",
                transactions=[{"type": "abandon", "value": True}],
            ),
            mock.call(
                rev_id="PHID-DREV-125",
                transactions=[{"type": "abandon", "value": True}],
            ),
        ]
        mock_conduit.apply_transactions_to_revision.assert_has_calls(expected_calls)

    @mock.patch("mozphab.commands.abandon.conduit")
    def test_abandon_already_abandoned(self, mock_conduit):
        """Test handling of already abandoned revisions."""
        # Setup mocks
        mock_conduit.check.return_value = True
        mock_conduit.get_revisions.return_value = [
            self.mock_revisions[1]
        ]  # Already abandoned

        # Call function
        with self.assertLogs() as logging_watcher:
            abandon.abandon_revisions([124], self.args)

        # Verify no abandonment calls were made
        mock_conduit.apply_transactions_to_revision.assert_not_called()

        # Verify warning was logged
        self.assertIn(
            Contains("All specified revisions are already abandoned."),
            logging_watcher.output,
        )

    @mock.patch("mozphab.commands.abandon.conduit")
    def test_abandon_connection_failure(self, mock_conduit):
        """Test handling of connection failures."""
        mock_conduit.check.return_value = False

        with self.assertRaises(exceptions.Error) as cm:
            abandon.abandon_revisions([123], self.args)

        self.assertIn("Failed to use Conduit API", str(cm.exception))

    @mock.patch("mozphab.commands.abandon.conduit")
    def test_abandon_revisions_not_found(self, mock_conduit):
        """Test handling of revisions not found."""
        mock_conduit.check.return_value = True
        mock_conduit.get_revisions.return_value = []

        with self.assertRaises(exceptions.Error) as cm:
            abandon.abandon_revisions([999], self.args)

        self.assertIn("No revisions found", str(cm.exception))

    @mock.patch("mozphab.commands.abandon.conduit")
    @mock.patch("mozphab.commands.abandon.prompt")
    def test_abandon_with_confirmation_no(self, mock_prompt, mock_conduit):
        """Test abandonment with confirmation prompt answering No."""
        # Setup mocks
        mock_conduit.check.return_value = True
        mock_conduit.get_revisions.return_value = [self.mock_revisions[0]]
        mock_prompt.return_value = "No"

        # Create args without --yes flag
        args = argparse.Namespace(yes=False)

        # Call function
        abandon.abandon_revisions([123], args)

        # Verify prompt was called
        mock_prompt.assert_called_once_with("Abandon these revisions?", ["Yes", "No"])

        # Verify no abandonment calls were made
        mock_conduit.apply_transactions_to_revision.assert_not_called()

    @mock.patch("mozphab.commands.abandon.conduit")
    @mock.patch("mozphab.commands.abandon.prompt")
    def test_abandon_with_confirmation_yes(self, mock_prompt, mock_conduit):
        """Test abandonment with confirmation prompt answering Yes."""
        # Setup mocks
        mock_conduit.check.return_value = True
        mock_conduit.get_revisions.return_value = [self.mock_revisions[0]]
        mock_prompt.return_value = "Yes"

        # Create args without --yes flag
        args = argparse.Namespace(yes=False)

        # Call function
        abandon.abandon_revisions([123], args)

        # Verify prompt was called
        mock_prompt.assert_called_once_with("Abandon these revisions?", ["Yes", "No"])

        # Verify abandonment call was made
        mock_conduit.apply_transactions_to_revision.assert_called_once_with(
            rev_id="PHID-DREV-123", transactions=[{"type": "abandon", "value": True}]
        )

    @mock.patch("mozphab.commands.abandon.conduit")
    def test_abandon_mixed_statuses(self, mock_conduit):
        """Test abandonment with mixed revision statuses."""
        # Setup mocks
        mock_conduit.check.return_value = True
        mock_conduit.get_revisions.return_value = self.mock_revisions  # Mix of statuses

        # Call function
        with self.assertLogs() as logging_watcher:
            abandon.abandon_revisions([123, 124, 125], self.args)

        # Verify only non-abandoned revisions were processed
        expected_calls = [
            mock.call(
                rev_id="PHID-DREV-123",
                transactions=[{"type": "abandon", "value": True}],
            ),
            mock.call(
                rev_id="PHID-DREV-125",
                transactions=[{"type": "abandon", "value": True}],
            ),
        ]
        mock_conduit.apply_transactions_to_revision.assert_has_calls(expected_calls)

        # Verify already abandoned message was logged
        self.assertIn(Contains("Already abandoned: D124"), logging_watcher.output)


class TestAbandonMainFunction(unittest.TestCase):
    """Test the main abandon function."""

    @mock.patch("mozphab.commands.abandon.abandon_revisions")
    def test_abandon_main_success(self, mock_abandon_revisions):
        """Test successful execution of main abandon function."""
        # Setup args with revision IDs (parsed by argparse)
        args = argparse.Namespace(revisions=[123, 124])

        # Call function
        abandon.abandon(None, args)

        # Verify calls
        mock_abandon_revisions.assert_called_once_with([123, 124], args)

    @mock.patch("mozphab.commands.abandon.abandon_revisions")
    def test_abandon_main_error_propagation(self, mock_abandon_revisions):
        """Test that errors are properly propagated."""
        # Setup mock to raise error
        mock_abandon_revisions.side_effect = exceptions.Error("Test error")
        args = argparse.Namespace(revisions=[123])

        # Verify error is propagated
        with self.assertRaises(exceptions.Error) as cm:
            abandon.abandon(None, args)

        self.assertEqual(str(cm.exception), "Test error")


class TestAbandonParser(unittest.TestCase):
    """Test the argument parser setup."""

    def test_add_parser(self):
        """Test that the parser is configured correctly."""
        import argparse

        # Create a parent parser to add our command to
        parent_parser = argparse.ArgumentParser()
        subparsers = parent_parser.add_subparsers()

        # Add our abandon parser
        abandon.add_parser(subparsers)

        # Test parsing valid arguments (check_revision_id converts to int)
        args = parent_parser.parse_args(["abandon", "D123", "D124"])
        self.assertEqual(args.revisions, [123, 124])
        self.assertFalse(args.yes)

        # Test with --yes flag
        args = parent_parser.parse_args(["abandon", "--yes", "D123"])
        self.assertEqual(args.revisions, [123])
        self.assertTrue(args.yes)

        # Test with -y flag
        args = parent_parser.parse_args(["abandon", "-y", "D123"])
        self.assertEqual(args.revisions, [123])
        self.assertTrue(args.yes)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import platform

import pytest

from mozphab.config import Config


def test_defaults(config):
    git_cmd = ["git.exe"] if platform.system() == "Windows" else ["git"]
    hg_cmd = ["hg.exe"] if platform.system() == "Windows" else ["hg"]

    assert config.no_ansi is False
    assert config.safe_mode is False
    assert config.git_remote == []
    assert config.git_command == git_cmd
    assert config.hg_command == hg_cmd
    assert config.auto_submit is False
    assert config.always_blocking is False
    assert config.warn_untracked is True
    assert config.apply_patch_to == "base"
    assert config.create_bookmark is True
    assert config.create_topic is False
    assert config.create_branch is True
    assert config.always_full_stack is False
    assert config.create_commit is True
    assert config.self_last_check == 0
    assert config.self_auto_update is True
    assert config.get_pre_releases is False
    assert config.report_to_sentry is True
    assert config.telemetry_enabled is False


def test_write(config):
    # Note: some of these values will NOT be overridden
    # based on how `config.write` is set up
    config.no_ansi = True
    config.safe_mode = True
    config.git_remote = ["test"]
    config.git_command = ["test"]
    config.hg_command = ["test"]
    config.auto_submit = True
    config.always_blocking = True
    config.warn_untracked = False
    config.apply_patch_to = "here"
    config.create_bookmark = False
    config.create_topic = True
    config.always_full_stack = True
    config.self_last_check = 12
    config.self_auto_update = False
    config.get_pre_releases = True
    config.report_to_sentry = False
    config.telemetry_enabled = True
    config.write()

    new_config = Config(filename=config.filename)
    git_cmd = ["git.exe"] if platform.system() == "Windows" else ["git"]

    assert new_config.no_ansi is True
    assert new_config.safe_mode is True
    assert new_config.git_remote[0] == "test"
    assert new_config.git_command[0] == git_cmd[0]
    assert new_config.hg_command[0] == "test"
    assert new_config.auto_submit is True
    assert new_config.always_blocking is True
    assert new_config.warn_untracked is False
    assert new_config.apply_patch_to == "here"
    assert new_config.create_bookmark is False
    assert new_config.create_topic is True
    assert new_config.always_full_stack is True
    assert new_config.self_last_check == 0
    assert new_config.self_auto_update is True
    assert new_config.get_pre_releases is False
    assert new_config.report_to_sentry is True
    assert new_config.telemetry_enabled is True


def test_invalid_int_field(config):
    config._set("updater", "self_last_check", "")
    config.write()

    with pytest.raises(ValueError) as e:
        Config(filename=config.filename)

    assert str(e.value) == (
        "could not convert updater.self_last_check to an integer: "
        "invalid literal for int() with base 10: ''"
    )


def test_invalid_bool_field(config):
    config.no_ansi = "test"
    config.write()

    with pytest.raises(ValueError) as e:
        Config(filename=config.filename)

    assert (
        str(e.value) == "could not convert ui.no_ansi to a boolean: Not a boolean: test"
    )



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import logging
from unittest import mock

import pytest
from callee import Contains

from mozphab import exceptions, mozphab
from mozphab.commands import reorganise
from mozphab.commits import Commit


@pytest.mark.parametrize(
    "phids,transactions,abandoned",
    [
        # No change
        (("A", "A"), {}, set()),
        (("ABC", "ABC"), {}, set()),
        (([], ["A"]), {}, set()),
        # Abandon
        (
            ("ABC", "A"),
            {
                "A": [{"type": "children.remove", "value": ["B"]}],
                "B": [
                    {"type": "children.remove", "value": ["C"]},
                    {"type": "abandon", "value": True},
                ],
                "C": [{"type": "abandon", "value": True}],
            },
            set(),
        ),
        (
            ("ABC", "B"),
            {
                "A": [
                    {"type": "children.remove", "value": ["B"]},
                    {"type": "abandon", "value": True},
                ],
                "B": [{"type": "children.remove", "value": ["C"]}],
                "C": [{"type": "abandon", "value": True}],
            },
            set(),
        ),
        (
            ("ABC", "C"),
            {
                "A": [
                    {"type": "children.remove", "value": ["B"]},
                    {"type": "abandon", "value": True},
                ],
                "B": [
                    {"type": "children.remove", "value": ["C"]},
                    {"type": "abandon", "value": True},
                ],
            },
            set(),
        ),
        (
            ("ABC", "C"),
            {
                "A": [
                    {"type": "children.remove", "value": ["B"]},
                ],
                "B": [
                    {"type": "children.remove", "value": ["C"]},
                    {"type": "abandon", "value": True},
                ],
            },
            {"A"},
        ),
        # Reorder
        (
            ("AB", "BA"),
            {
                "A": [{"type": "children.remove", "value": ["B"]}],
                "B": [{"type": "children.set", "value": ["A"]}],
            },
            set(),
        ),
        (
            ("ABC", "BC"),
            {
                "A": [
                    {"type": "children.remove", "value": ["B"]},
                    {"type": "abandon", "value": True},
                ]
            },
            set(),
        ),
        (
            ("ABC", "ACB"),
            {
                "A": [{"type": "children.set", "value": ["C"]}],
                "B": [{"type": "children.remove", "value": ["C"]}],
                "C": [{"type": "children.set", "value": ["B"]}],
            },
            set(),
        ),
        (
            ("ABC", "BAC"),
            {
                "A": [{"type": "children.set", "value": ["C"]}],
                "B": [{"type": "children.set", "value": ["A"]}],
            },
            set(),
        ),
        (
            ("ABC", "CAB"),
            {
                "B": [{"type": "children.remove", "value": ["C"]}],
                "C": [{"type": "children.set", "value": ["A"]}],
            },
            set(),
        ),
        (
            ("ABC", "CBA"),
            {
                "A": [{"type": "children.remove", "value": ["B"]}],
                "B": [{"type": "children.set", "value": ["A"]}],
                "C": [{"type": "children.set", "value": ["B"]}],
            },
            set(),
        ),
        # Insert
        (("ABC", "DABC"), {"D": [{"type": "children.set", "value": ["A"]}]}, set()),
        (
            ("ABC", "ADBC"),
            {
                "A": [{"type": "children.set", "value": ["D"]}],
                "D": [{"type": "children.set", "value": ["B"]}],
            },
            set(),
        ),
        (
            ("ABC", "ABDC"),
            {
                "B": [{"type": "children.set", "value": ["D"]}],
                "D": [{"type": "children.set", "value": ["C"]}],
            },
            set(),
        ),
        (
            ("ABC", "BCAD"),
            {
                "A": [{"type": "children.set", "value": ["D"]}],
                "C": [{"type": "children.set", "value": ["A"]}],
            },
            set(),
        ),
        (([], ["A", "B"]), {"A": [{"type": "children.set", "value": ["B"]}]}, set()),
        # Insert and reorder
        (
            ("ABC", "DCAB"),
            {
                "B": [{"type": "children.remove", "value": ["C"]}],
                "C": [{"type": "children.set", "value": ["A"]}],
                "D": [{"type": "children.set", "value": ["C"]}],
            },
            set(),
        ),
        (
            ("ABC", "CDAB"),
            {
                "B": [{"type": "children.remove", "value": ["C"]}],
                "C": [{"type": "children.set", "value": ["D"]}],
                "D": [{"type": "children.set", "value": ["A"]}],
            },
            set(),
        ),
        (
            ("ABC", "CADB"),
            {
                "A": [{"type": "children.set", "value": ["D"]}],
                "B": [{"type": "children.remove", "value": ["C"]}],
                "C": [{"type": "children.set", "value": ["A"]}],
                "D": [{"type": "children.set", "value": ["B"]}],
            },
            set(),
        ),
        (
            ("ABC", "CABD"),
            {
                "B": [{"type": "children.set", "value": ["D"]}],
                "C": [{"type": "children.set", "value": ["A"]}],
            },
            set(),
        ),
        # Nothing in common
        (("A", "B"), {"A": [{"type": "abandon", "value": True}]}, set()),
    ],
)
def test_stack_transactions_with_abandon(phids, transactions, abandoned):
    remote_phids, local_phids = phids
    assert (
        reorganise.stack_transactions(remote_phids, local_phids, abandoned)
        == transactions
    )


@pytest.mark.parametrize(
    "phids,transactions,abandoned",
    [
        (
            ("ABC", "A"),
            {
                "A": [{"type": "children.remove", "value": ["B"]}],
                "B": [
                    {"type": "children.remove", "value": ["C"]},
                ],
            },
            set(),
        ),
        (
            ("ABC", "B"),
            {
                "A": [
                    {"type": "children.remove", "value": ["B"]},
                ],
                "B": [{"type": "children.remove", "value": ["C"]}],
            },
            set(),
        ),
        (
            ("ABC", "C"),
            {
                "A": [
                    {"type": "children.remove", "value": ["B"]},
                ],
                "B": [
                    {"type": "children.remove", "value": ["C"]},
                ],
            },
            set(),
        ),
        (
            ("ABC", "C"),
            {
                "A": [
                    {"type": "children.remove", "value": ["B"]},
                ],
                "B": [
                    {"type": "children.remove", "value": ["C"]},
                ],
            },
            {"A"},
        ),
        (
            ("ABC", "BC"),
            {
                "A": [
                    {"type": "children.remove", "value": ["B"]},
                ]
            },
            set(),
        ),
        # Nothing in common
        (("A", "B"), {}, set()),
    ],
)
def test_stack_transactions_no_abandon(phids, transactions, abandoned):
    remote_phids, local_phids = phids
    assert (
        reorganise.stack_transactions(
            remote_phids, local_phids, abandoned, no_abandon=True
        )
        == transactions
    )


@pytest.mark.parametrize(
    "stacks,expected",
    [
        (
            ({"A": None}, [Commit(rev_id=1)], ["A"]),
            (["A"], ["A"], set(), {"no_abandon": False}),
        ),
        (
            ({"B": None}, [Commit(rev_id=1)], ["A"]),
            (["B"], ["A"], set(), {"no_abandon": False}),
        ),
        (
            ({"A": "B", "B": None}, [Commit(rev_id=1)], ["A"]),
            (["A", "B"], ["A"], set(), {"no_abandon": False}),
        ),
        (
            (
                {"A": None},
                [Commit(rev_id=1), Commit(rev_id=2)],
                ["A", "B"],
            ),
            (["A"], ["A", "B"], set(), {"no_abandon": False}),
        ),
        (
            ({"A": None}, [Commit(rev_id=1), Commit(rev_id=2)], ["A", "B"]),
            (["A"], ["A", "B"], set(), {"no_abandon": True}),
        ),
    ],
)
@mock.patch("mozphab.commands.reorganise.stack_transactions")
@mock.patch("mozphab.conduit.ConduitAPI.check")
@mock.patch("mozphab.commands.reorganise.augment_commits_from_body")
@mock.patch("mozphab.commands.reorganise.convert_stackgraph_to_linear")
@mock.patch("mozphab.conduit.ConduitAPI.get_revisions")
@mock.patch("mozphab.conduit.ConduitAPI.ids_to_phids")
@mock.patch("mozphab.conduit.ConduitAPI.phids_to_ids")
@mock.patch("mozphab.conduit.ConduitAPI.edit_revision")
def test_reorg_calling_stack_transactions(
    _edit_revision,
    _phid2id,
    m_id2phid,
    _get_revs,
    m_convert_stackgraph_to_linear,
    _augment_commits,
    _check,
    m_trans,
    git,
    stacks,
    expected,
):
    *args, kwargs = expected

    class Args:
        yes = True
        no_abandon = kwargs["no_abandon"]
        force = False
        no_abandon_unconnected = False
        no_hyperlinks = False
        verbose = False

    phabstack, commits, rev_ids = stacks
    m_convert_stackgraph_to_linear.return_value = phabstack
    mozphab.conduit.set_repo(git)
    mozphab.conduit.repo.commit_stack = mock.Mock()
    mozphab.conduit.repo.commit_stack.return_value = commits
    m_id2phid.return_value = rev_ids
    reorganise.reorganise(git, Args())
    m_trans.assert_called_once_with(*args, **kwargs)


@mock.patch("mozphab.conduit.ConduitAPI.check")
def test_conduit_broken(m_check):
    m_check.return_value = False

    class Args:
        force = False
        no_abandon_unconnected = False
        no_hyperlinks = False
        verbose = False

    with pytest.raises(exceptions.Error) as e:
        reorganise.reorganise(None, Args())

    assert str(e.value) == "Failed to use Conduit API"


@mock.patch("mozphab.conduit.ConduitAPI.check")
@mock.patch("mozphab.commands.reorganise.augment_commits_from_body")
def test_commits_invalid(_augment, _check, git):
    class Args:
        force = False
        no_abandon_unconnected = False
        no_hyperlinks = False
        verbose = False

    mozphab.conduit.set_repo(git)
    mozphab.conduit.repo.commit_stack = mock.Mock()
    mozphab.conduit.repo.commit_stack.return_value = []
    with pytest.raises(exceptions.Error) as e:
        reorganise.reorganise(git, Args())

    assert str(e.value) == "Failed to find any commits to reorganise."

    mozphab.conduit.repo.commit_stack.return_value = [Commit(rev_id=None, name="A")]
    with pytest.raises(exceptions.Error) as e:
        reorganise.reorganise(git, Args())

    error = str(e.value)
    assert error.startswith("Found new commit in the local stack: A.")


@mock.patch("mozphab.conduit.ConduitAPI.phids_to_ids")
@mock.patch("mozphab.commands.reorganise.walk_llist")
def test_remote_stack_invalid(m_walk, m_ids, git, caplog: pytest.LogCaptureFixture):
    class Args:
        force = False
        no_abandon_unconnected = False
        no_hyperlinks = False
        verbose = False

    caplog.set_level(logging.ERROR)
    m_ids.return_value = ["A", "B", "C"]
    m_walk.side_effect = exceptions.Error("TEST")

    with (
        mock.patch("mozphab.conduit.ConduitAPI.check"),
        mock.patch("mozphab.git.Git.commit_stack"),
        mock.patch("mozphab.conduit.ConduitAPI.get_revisions"),
        mock.patch("mozphab.commands.reorganise.convert_stackgraph_to_linear"),
        pytest.raises(exceptions.Error, match="TEST"),
    ):
        reorganise.reorganise(git, Args())

    assert caplog.messages == [
        "Remote stack is not linear.\nDetected stack:\nA <- B <- C"
    ]


@pytest.mark.parametrize(
    "llist,error_text",
    (
        ({"a": "b", "c": None}, "Multiple heads"),
        ({"a": "b", "b": None, "c": "d", "d": None}, "Multiple heads"),
        ({}, "Failed to find head"),
        ({"a": "b", "b": "c", "c": "a"}, "Failed to find head"),
        ({"a": "b", "b": "c", "c": "b"}, "Dependency loop"),
    ),
)
def test_walk_llist_errors(llist, error_text):
    walk = reorganise.walk_llist
    with pytest.raises(exceptions.Error) as e:
        walk(llist)

    assert str(e.value).startswith(error_text)


@pytest.mark.parametrize(
    "llist,nodes",
    (
        ({"a": None}, ["a"]),
        ({"a": "b", "b": None}, ["a", "b"]),
        ({"a": "b"}, ["a", "b"]),
        ({"b": "a", "a": None}, ["b", "a"]),
        ({"a": "b", "b": "c", "c": None}, ["a", "b", "c"]),
    ),
)
def test_walk_llist(llist, nodes):
    assert reorganise.walk_llist(llist) == nodes


@pytest.mark.parametrize(
    "params,result",
    (
        ([], {}),
        (["A"], {"A": None}),
        (["A", "B"], {"A": "B", "B": None}),
        (["A", "A"], {"A": None}),
    ),
)
def test_to_llist(params, result):
    assert reorganise.to_llist(params) == result


@pytest.mark.parametrize(
    "stack_graph,phabstack,message",
    (
        ({"1": []}, {}, "Single node should produce no stack."),
        ({"1": [], "2": []}, {}, "Multiple single nodes should produce no stack."),
        (
            {"1": ["2"], "2": []},
            {"2": "1", "1": None},
            "Two elements should be linked together.",
        ),
        (
            {"1": ["2"], "2": [], "3": []},
            {"2": "1", "1": None},
            "Two elements should be linked together and extra single node is ignored.",
        ),
        # Multiple parents.
        (
            {"1": ["2", "3"], "2": [], "3": []},
            {"3": "1", "2": "1", "1": None},
            "Multiple parents should be allowed.",
        ),
        (
            {"1": ["2", "3"], "2": ["5"], "3": ["4"], "4": [], "5": []},
            {"4": "3", "5": "2", "3": "1", "2": "1", "1": None},
            "Multiple parents should be allowed.",
        ),
    ),
)
def test_convert_stackgraph_to_linear_pass(stack_graph, phabstack, message):
    phids_to_ids = {phid: int(phid) for phid in stack_graph}
    assert (
        reorganise.convert_stackgraph_to_linear(stack_graph, phids_to_ids) == phabstack
    ), message


@pytest.mark.parametrize(
    "stack_graph",
    (
        # Multiple children.
        {"1": ["2"], "2": [], "3": ["2"]},
        ({"1": ["2", "3"], "2": [], "3": ["2"]}),
    ),
)
def test_convert_stackgraph_to_linear_fail(stack_graph):
    phids_to_ids = {phid: int(phid) for phid in stack_graph}
    with pytest.raises(exceptions.Error):
        reorganise.convert_stackgraph_to_linear(stack_graph, phids_to_ids)


@pytest.mark.parametrize(
    "remote_phids,local_phids,abandoned,expected_transactions",
    [
        # Simple case: same revision in both remote and local
        (
            ["A"],
            ["A"],
            set(),
            {
                "A": [
                    {"type": "children.set", "value": []},
                ]
            },
        ),
        # Local revision with child
        (
            ["A"],
            ["A", "B"],
            set(),
            {
                "A": [
                    {"type": "children.set", "value": []},
                    {"type": "children.set", "value": ["B"]},
                ]
            },
        ),
        # Remote revision not in local - should be abandoned
        (
            ["A", "B"],
            ["A"],
            set(),
            {
                "A": [
                    {"type": "children.set", "value": []},
                ],
                "B": [
                    {"type": "abandon", "value": True},
                ],
            },
        ),
        # Remote revision not in local but already abandoned - no abandon transaction
        (
            ["A", "B"],
            ["A"],
            {"B"},
            {
                "A": [
                    {"type": "children.set", "value": []},
                ]
            },
        ),
        # Complex case: reorder with abandonment
        (
            ["A", "B", "C"],
            ["C", "A"],
            set(),
            {
                "A": [
                    {"type": "children.set", "value": []},
                ],
                "B": [
                    {"type": "abandon", "value": True},
                ],
                "C": [
                    {"type": "children.set", "value": []},
                    {"type": "children.set", "value": ["A"]},
                ],
            },
        ),
        # Empty remote, non-empty local
        (
            [],
            ["A", "B"],
            set(),
            {
                "A": [
                    {"type": "children.set", "value": ["B"]},
                ]
            },
        ),
        # Empty local, non-empty remote - abandon all
        (
            ["A", "B"],
            [],
            set(),
            {
                "A": [
                    {"type": "abandon", "value": True},
                ],
                "B": [
                    {"type": "abandon", "value": True},
                ],
            },
        ),
    ],
)
def test_force_stack_transactions(
    remote_phids, local_phids, abandoned, expected_transactions
):
    """Test the force_stack_transactions function with various scenarios."""
    result = reorganise.force_stack_transactions(remote_phids, local_phids, abandoned)
    assert result == expected_transactions


@pytest.mark.parametrize(
    "remote_phids,local_phids,abandoned,no_abandon_unconnected,expected_transactions",
    [
        # With no_abandon_unconnected=True: don't abandon remote revisions not in local
        (
            ["A", "B", "C"],
            ["A"],
            set(),
            True,
            {
                "A": [
                    {"type": "children.set", "value": []},
                ],
                "B": [
                    {"type": "children.set", "value": []},
                ],
                "C": [
                    {"type": "children.set", "value": []},
                ],
            },
        ),
        # With no_abandon_unconnected=False: abandon remote revisions not in local (default behavior)
        (
            ["A", "B", "C"],
            ["A"],
            set(),
            False,
            {
                "A": [
                    {"type": "children.set", "value": []},
                ],
                "B": [
                    {"type": "abandon", "value": True},
                ],
                "C": [
                    {"type": "abandon", "value": True},
                ],
            },
        ),
        # With no_abandon_unconnected=True and already abandoned revision: no change
        (
            ["A", "B"],
            ["A"],
            {"B"},
            True,
            {
                "A": [
                    {"type": "children.set", "value": []},
                ],
                "B": [
                    {"type": "children.set", "value": []},
                ],
            },
        ),
        # Complex case with no_abandon_unconnected=True: reorder but don't abandon
        (
            ["A", "B", "C"],
            ["C", "A"],
            set(),
            True,
            {
                "A": [
                    {"type": "children.set", "value": []},
                ],
                "B": [
                    {"type": "children.set", "value": []},
                ],
                "C": [
                    {"type": "children.set", "value": []},
                    {"type": "children.set", "value": ["A"]},
                ],
            },
        ),
    ],
)
def test_force_stack_transactions_no_abandon_unconnected(
    remote_phids, local_phids, abandoned, no_abandon_unconnected, expected_transactions
):
    """Test the force_stack_transactions function with no_abandon_unconnected flag."""
    result = reorganise.force_stack_transactions(
        remote_phids, local_phids, abandoned, no_abandon_unconnected
    )
    assert result == expected_transactions


@pytest.mark.parametrize(
    "stacks,expected_args",
    [
        # Force mode with all local revisions present
        (
            ({"A": None}, [Commit(rev_id=1)], ["A"], True),
            (["A"], ["A"], set()),
        ),
        # Force mode with reordering
        (
            (
                {"A": "B", "B": None},
                [Commit(rev_id=2), Commit(rev_id=1)],
                ["B", "A"],
                True,
            ),
            (["A", "B"], ["B", "A"], set()),
        ),
    ],
)
@mock.patch("mozphab.commands.reorganise.force_stack_transactions")
@mock.patch("mozphab.commands.reorganise.stack_transactions")
@mock.patch("mozphab.conduit.ConduitAPI.check")
@mock.patch("mozphab.commands.reorganise.augment_commits_from_body")
@mock.patch("mozphab.commands.reorganise.convert_stackgraph_to_linear")
@mock.patch("mozphab.conduit.ConduitAPI.get_revisions")
@mock.patch("mozphab.conduit.ConduitAPI.ids_to_phids")
@mock.patch("mozphab.conduit.ConduitAPI.phids_to_ids")
@mock.patch("mozphab.conduit.ConduitAPI.apply_transactions_to_revision")
def test_reorg_force_mode(
    _apply_transactions,
    _phid2id,
    m_id2phid,
    _get_revs,
    m_convert_stackgraph_to_linear,
    _augment_commits,
    _check,
    m_stack_trans,
    m_force_trans,
    git,
    stacks,
    expected_args,
):
    """Test that force mode calls force_stack_transactions instead of stack_transactions."""
    phabstack, commits, rev_ids, force_mode = stacks

    class Args:
        yes = True
        no_abandon = False
        force = force_mode
        no_abandon_unconnected = False
        no_hyperlinks = False
        verbose = False

    m_convert_stackgraph_to_linear.return_value = phabstack
    mozphab.conduit.set_repo(git)
    mozphab.conduit.repo.commit_stack = mock.Mock()
    mozphab.conduit.repo.commit_stack.return_value = commits
    m_id2phid.return_value = rev_ids

    # Mock return values for transactions
    m_force_trans.return_value = {"A": [{"type": "children.set", "value": []}]}
    m_stack_trans.return_value = {"A": [{"type": "children.set", "value": []}]}

    reorganise.reorganise(git, Args())

    if force_mode:
        m_force_trans.assert_called_once_with(
            *expected_args, no_abandon_unconnected=False
        )
        m_stack_trans.assert_not_called()
    else:
        m_stack_trans.assert_called_once()
        m_force_trans.assert_not_called()


@mock.patch("mozphab.conduit.ConduitAPI.check")
@mock.patch("mozphab.commands.reorganise.augment_commits_from_body")
def test_force_mode_requires_all_local_revisions_on_phabricator(_augment, _check, git):
    """Test that force mode requires all local revisions to be present on Phabricator."""

    class Args:
        force = True
        no_abandon_unconnected = False
        no_hyperlinks = False
        verbose = False

    mozphab.conduit.set_repo(git)
    mozphab.conduit.repo.commit_stack = mock.Mock()
    mozphab.conduit.repo.commit_stack.return_value = [Commit(rev_id=None, name="A")]

    with pytest.raises(exceptions.Error) as e:
        reorganise.reorganise(git, Args())

    assert (
        str(e.value)
        == "Force mode requires all local revisions to be present on Phabricator."
    )


@mock.patch("mozphab.conduit.ConduitAPI.check")
def test_no_abandon_unconnected_requires_force(_check):
    """Test that --no-abandon-unconnected requires --force flag."""

    class Args:
        force = False
        no_abandon_unconnected = True
        no_hyperlinks = False
        verbose = False

    _check.return_value = True

    with pytest.raises(exceptions.Error) as e:
        reorganise.reorganise(None, Args())

    assert str(e.value) == "--no-abandon-unconnected can only be used with --force"


@mock.patch("mozphab.commands.reorganise.walk_llist")
def test_force_mode_ignores_remote_stack_errors(
    m_walk, git, caplog: pytest.LogCaptureFixture
):
    """Test that force mode ignores remote stack structure issues."""

    class Args:
        force = True
        yes = True
        no_abandon_unconnected = False
        no_hyperlinks = False
        verbose = False

    caplog.set_level(logging.WARNING, logger="moz-phab")
    m_walk.side_effect = exceptions.Error("Remote stack is not linear")

    # Mock the conduit methods that would be called
    with (
        mock.patch("mozphab.conduit.ConduitAPI.check"),
        mock.patch("mozphab.git.Git.commit_stack"),
        mock.patch("mozphab.conduit.ConduitAPI.get_revisions"),
        mock.patch("mozphab.commands.reorganise.convert_stackgraph_to_linear"),
    ):
        # This should not raise an error in force mode
        reorganise.reorganise(git, Args())

    # Should log a warning instead of an error
    # Check that the specific warning was called
    for record in caplog.records:
        assert record.levelno == logging.WARNING
    assert (
        Contains("Remote stack is not linear, but continuing in force mode")
        in caplog.messages
    )


class TestReorganiseParser:
    """Test the argument parser setup for reorganise command."""

    def test_add_parser(self):
        """Test that the parser is configured correctly for the --force flag."""
        import argparse

        # Create a parent parser to add our command to
        parent_parser = argparse.ArgumentParser()
        subparsers = parent_parser.add_subparsers()

        # Add our reorganise parser
        reorganise.add_parser(subparsers)

        # Test parsing without --force flag
        args = parent_parser.parse_args(["reorg"])
        assert args.force is False
        assert args.yes is False
        assert args.no_abandon is False

        # Test parsing with --force flag
        args = parent_parser.parse_args(["reorg", "--force"])
        assert args.force is True
        assert args.yes is False
        assert args.no_abandon is False

        # Test parsing with multiple flags
        args = parent_parser.parse_args(["reorg", "--force", "--yes", "--no-abandon"])
        assert args.force is True
        assert args.yes is True
        assert args.no_abandon is True

        # Test parsing with revision arguments
        args = parent_parser.parse_args(["reorg", "--force", "abc123", "def456"])
        assert args.force is True
        assert args.start_rev == "abc123"
        assert args.end_rev == "def456"

        # Test parsing with upstream argument
        args = parent_parser.parse_args(["reorg", "--force", "--upstream", "main"])
        assert args.force is True
        assert args.upstream == ["main"]

        # Test parsing with --no-abandon-unconnected flag
        args = parent_parser.parse_args(
            ["reorg", "--force", "--no-abandon-unconnected"]
        )
        assert args.force is True
        assert args.no_abandon_unconnected is True

        # Test parsing without --no-abandon-unconnected flag
        args = parent_parser.parse_args(["reorg", "--force"])
        assert args.force is True
        assert args.no_abandon_unconnected is False



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from typing import Any


class SimpleCache:
    """Simple key/value store with all lowercase keys."""

    def __init__(self):
        self._cache = {}

    def __contains__(self, key: str) -> bool:
        return key.lower() in self._cache

    def get(self, key: str) -> Any:
        return self._cache.get(key.lower())

    def set(self, key: str, value: Any):
        self._cache[key.lower()] = value

    def delete(self, key: str):
        if key in self:
            del self._cache[key.lower()]

    def reset(self):
        self._cache = {}


cache = SimpleCache()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import argparse
import re
import subprocess
from typing import (
    List,
    Optional,
    Tuple,
)

from mozphab.conduit import conduit
from mozphab.config import (
    Config,
    config,
)
from mozphab.exceptions import Error, NonLinearException, NotFoundError
from mozphab.helpers import prepare_body, prompt, short_node
from mozphab.logger import logger
from mozphab.mercurial import Mercurial
from mozphab.patch import apply_patch
from mozphab.repository import (
    Repository,
)
from mozphab.spinner import wait_message


def get_base_ref(diff: dict) -> Optional[str]:
    """Given a diff, return the base revision SHA the diff was based on."""
    for ref in diff["fields"].get("refs", []):
        if ref["type"] == "base":
            return ref["identifier"]


def get_diff_by_id(diff_id: int) -> Tuple[str, dict]:
    """Retrieves a diff from Phabricator

    Args:
    * diff_id: the ID of the diff to retrieve

    Returns:
    * a tuple containing:
        * the PHID of the requested diff
        * the dictionary returned by Phabricator
    """
    diff_dict = conduit.get_diffs(ids=[diff_id])
    if not diff_dict:
        raise NotFoundError(f"Could not find diff with ID of {diff_id}.")

    if len(diff_dict) != 1:
        raise Error(f"Unexpected result received from Phabricator for Diff {diff_id}.")

    requested_diff_phid = list(diff_dict.keys())[0]
    return requested_diff_phid, diff_dict[requested_diff_phid]


def update_revision_with_new_diff(revs: List[dict], diff: dict) -> None:
    """Updates the revision to point to the given diff if they are related

    Args:
    * revs: list of revisions
    * diff: the new diff to point to

    Returns:
    * None if the diff is related to a revision in the list

    Raises:
    * Error if no relation is found
    """
    for rev in revs:
        if diff["fields"]["revisionPHID"] == rev["phid"]:
            rev["fields"]["diffPHID"] = diff["phid"]
            return

    raise Error(f"Diff {diff['id']} is not related to any revision in the stack.")


def resolve_branch_name(
    args: argparse.Namespace, config: Config, rev_id: str
) -> Optional[str]:
    """Resolve the branch name for the resulting patch.

    Use the value passed from `--name` on the CLI if possible. If
    `--no-commit` is passed, we don't need a branch name since we won't
    be committing to the VCS. Otherwise, format the patch from the
    `patch.branch_name_template` config knob.

    `patch.branch_name_template` supports a single format string, `rev_id`.
    """
    if args.name:
        # Return the value passed from the CLI.
        return args.name

    if args.no_commit or not config.create_commit:
        # `no_commit` implies no branch name.
        return None

    # Build the branch name from the configured template.
    return config.branch_name_template.format(rev_id=rev_id)


def patch(repo: Repository, args: argparse.Namespace):
    """Patch repository from Phabricator's revisions.

    By default:
    * perform sanity checks
    * find the base commit
    * create a new branch/bookmark/topic
    * apply the patches and commit the changes

    args.no_commit or config.create_commit is False - no commit will be created after applying diffs
    args.apply_to - <head|tip|branch> (default: branch)
        branch - find base commit and apply on top of it
        head/tip - apply changes to current commit
    args.raw is True - only print out the diffs (--force doesn't change anything)

    Raises:
    * Error if uncommitted changes are present in the working tree
    * Error if Phabricator revision is not found
    * Error if `--apply-to base` and no base commit found in the first diff
    * Error if base commit not found in repository
    * Error if `--diff-id` does not belong to any revision in the stack
    """
    # Check if raw Conduit API can be used
    with wait_message("Checking connection to Phabricator."):
        # Check if raw Conduit API can be used
        if not conduit.check():
            raise Error("Failed to use Conduit API")

    if not args.raw:
        # Check if local and remote VCS matches
        with wait_message("Checking VCS"):
            repo.check_vcs()

        # Look for any uncommitted changes
        with wait_message("Checking repository.."):
            clean = repo.is_worktree_clean()

        if not clean:
            raise Error(
                "Uncommitted changes present. Please %s them or commit before patching."
                % ("shelve" if isinstance(repo, Mercurial) else "stash")
            )

    # Get the target revision
    with wait_message("Fetching D%s.." % args.revision_id):
        revs = conduit.get_revisions(ids=[args.revision_id])

    if not revs:
        raise Error("Revision not found")

    revision = revs[0]

    if not args.skip_dependencies:
        with wait_message("Fetching D%s children.." % args.revision_id):
            try:
                children = conduit.get_successor_phids(
                    revision["phid"], include_abandoned=args.include_abandoned
                )
                non_linear = False
            except NonLinearException:
                children = []
                non_linear = True

        patch_children = True
        if children:
            if args.yes or config.always_full_stack:
                patch_children = True

            else:
                children_msg = (
                    "a child commit" if len(children) == 1 else "child commits"
                )
                res = prompt(
                    "Revision D%s has %s.  Would you like to patch the "
                    "full stack?." % (args.revision_id, children_msg),
                    ["Yes", "No", "Always"],
                )
                if res == "Always":
                    config.always_full_stack = True
                    config.write()

                patch_children = res == "Yes" or res == "Always"

            if patch_children:
                if non_linear and not args.yes:
                    logger.warning(
                        "Revision D%s has a non-linear successor graph.\n"
                        "Unable to apply the full stack.",
                        args.revision_id,
                    )
                    res = prompt("Continue with only part of the stack?", ["Yes", "No"])
                    if res == "No":
                        return

        # Get list of PHIDs in the stack
        try:
            with wait_message("Fetching D%s parents.." % args.revision_id):
                phids = conduit.get_ancestor_phids(revision["phid"])
        except NonLinearException:
            raise Error("Non linear dependency detected. Unable to patch the stack.")

        # Pull revisions data
        if phids:
            with wait_message("Fetching related revisions.."):
                revs.extend(conduit.get_revisions(phids=phids))
            revs.reverse()

        if children and patch_children:
            with wait_message("Fetching related revisions.."):
                revs.extend(conduit.get_revisions(phids=children))

    # Set the target id
    rev_id = revs[-1]["id"]

    logger.info(
        "Patching revision%s: %s",
        "s" if len(revs) > 1 else "",
        " ".join(["D%s" % r["id"] for r in revs]),
    )

    # Pull diffs
    with wait_message("Downloading patch information.."):
        diffs = conduit.get_diffs(phids=[r["fields"]["diffPHID"] for r in revs])

    # If a user specifies a diff ID, retrieve the diff and add it to the diff mapping,
    # and overwrite the diffPHID for the relevant revision
    if args.diff_id:
        requested_diff_phid, requested_diff = get_diff_by_id(args.diff_id)
        diffs[requested_diff_phid] = requested_diff
        update_revision_with_new_diff(revs, requested_diff)

    if not args.no_commit and config.create_commit and not args.raw:
        for rev in revs:
            diff = diffs[rev["fields"]["diffPHID"]]
            if not diff["attachments"]["commits"]["commits"]:
                raise Error(
                    "A diff without commit information detected in revision D%s.\n"
                    "Use `--no-commit` to patch the working tree." % rev["id"]
                )

    base_node = None
    if not args.raw:
        args.apply_to = args.apply_to or config.apply_patch_to

        if args.apply_to == "base":
            base_node = get_base_ref(diffs[revs[0]["fields"]["diffPHID"]])

            if not base_node:
                raise Error(
                    "Base commit not found in diff. "
                    "Use `--apply-to here` to patch current commit."
                )
        elif args.apply_to != "here":
            base_node = args.apply_to

        if args.apply_to != "here":
            try:
                with wait_message("Checking %s.." % short_node(base_node)):
                    base_node = repo.check_node(base_node)
            except NotFoundError as e:
                msg = "Unknown revision: %s" % short_node(base_node)
                if str(e):
                    msg += "\n%s" % str(e)

                if args.apply_to == "base":
                    msg += "\nUse --apply-to to set the base commit."

                raise Error(msg)

        branch_name = resolve_branch_name(args, config, rev_id)
        repo.before_patch(base_node, branch_name)

    for rev in revs:
        # Prepare the body using just the data from Phabricator
        body = prepare_body(
            rev["fields"]["title"],
            rev["fields"]["summary"],
            rev["id"],
            repo.phab_url,
        )
        diff = diffs[rev["fields"]["diffPHID"]]
        with wait_message("Downloading D%s.." % rev["id"]):
            raw = conduit.call("differential.getrawdiff", {"diffID": diff["id"]})

        if args.no_commit or not config.create_commit:
            with wait_message("Applying D%s.." % rev["id"]):
                apply_patch(raw, repo.path)
        else:
            try:
                diff_commits = diff["attachments"]["commits"]["commits"]
                author = "%s <%s>" % (
                    diff_commits[0]["author"]["name"],
                    diff_commits[0]["author"]["email"],
                )
            except (IndexError, KeyError):
                author = None
            try:
                date_created = diff["fields"]["dateCreated"]
            except KeyError:
                date_created = None

            if args.raw:
                # print rather than use logger.info; there's no need for this
                # to be in our logs.
                print(repo.format_patch(raw, body, author, date_created))

            else:
                try:
                    with wait_message("Applying D%s.." % rev["id"]):
                        repo.apply_patch(raw, body, author, date_created)
                except subprocess.CalledProcessError:
                    raise Error("Patch failed to apply")

        if rev["id"] != revs[-1]["id"]:
            logger.info("D%s applied", rev["id"])

    logger.warning("D%s applied", rev_id)


def check_revision_id(value: str) -> int:
    """Parse the revision ID from `value`.

    `value` is a `str` which is either `<id>`, `D<id>`,
    or a Phabricator revision URL.
    """
    # D123 or 123
    m = re.search(r"^D?(\d+)$", value)
    if m:
        return int(m.group(1))

    # Full URL
    m = re.search(r"^https?://[^/]+/D(\d+)", value)
    if m:
        return int(m.group(1))

    # Invalid
    raise argparse.ArgumentTypeError(
        "Invalid Revision ID (expected number or URL): %s\n" % value
    )


def add_parser(parser):
    patch_parser = parser.add_parser("patch", help="Patch from Phabricator revision.")
    patch_parser.add_argument(
        "revision_id", type=check_revision_id, help="Revision number."
    )

    # `--apply-to` and `--raw` are mutually exclusive.
    patch_group = patch_parser.add_mutually_exclusive_group()
    patch_group.add_argument(
        "--apply-to",
        "--applyto",
        "-a",
        metavar="TARGET",
        dest="apply_to",
        help="Where to apply the patch? <{NODE}|here|base> (default: %s)."
        % config.apply_patch_to,
    )
    patch_group.add_argument(
        "--raw", action="store_true", help="Prints out the raw diff to the STDOUT."
    )

    patch_parser.add_argument(
        "--diff-id",
        metavar="DIFF_ID",
        dest="diff_id",
        type=int,
        help="The ID of the diff to apply.",
    )
    patch_parser.add_argument(
        "--name",
        "-n",
        dest="name",
        metavar="NAME",
        help="Use the given name for the bookmark, topic, or branch.",
    )
    patch_parser.add_argument(
        "--no-commit",
        "--nocommit",
        action="store_true",
        dest="no_commit",
        help="Do not commit. Applies the changes with the `patch` command.",
    )
    patch_parser.add_argument(
        "--no-bookmark",
        "--nobookmark",
        action="store_true",
        dest="no_bookmark",
        help="(Mercurial only) Do not create the bookmark.",
    )
    patch_parser.add_argument(
        "--no-topic",
        "--notopic",
        action="store_true",
        dest="no_topic",
        help="(Mercurial only) Do not create the topic.",
    )
    patch_parser.add_argument(
        "--no-branch",
        "--nobranch",
        action="store_true",
        dest="no_branch",
        help="(Git only) Do not create the branch.",
    )
    patch_parser.add_argument(
        "--skip-dependencies",
        action="store_true",
        help="Do not search for dependencies; patch only one revision.",
    )
    patch_parser.add_argument(
        "--include-abandoned", action="store_true", help="Apply abandoned revisions."
    )
    patch_parser.add_argument(
        "--yes",
        "-y",
        action="store_true",
        help="Patch without confirmation (default: False).",
    )
    patch_parser.add_argument(
        "--safe-mode",
        dest="safe_mode",
        action="store_true",
        help="Run VCS with only necessary extensions.",
    )
    patch_parser.add_argument(
        "--force-vcs",
        action="store_true",
        help="EXPERIMENTAL: Override VCS compatibility check.",
    )
    patch_parser.set_defaults(func=patch, needs_repo=True)



#!/usr/bin/env python3
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# coding=utf-8

"""
CLI to support submission of a series of commits to Phabricator. .
"""

import argparse
import logging
import os
import ssl
import subprocess
import sys
import traceback
from typing import List

from packaging.version import Version

from mozphab import environment

from .args import parse_args
from .conduit import ConduitAPIError, conduit
from .config import config
from .detect_repository import repo_from_args
from .exceptions import Error
from .logger import init_logging, logger, stop_logging
from .repository import Repository
from .sentry import init_sentry, report_to_sentry
from .spinner import wait_message
from .telemetry import configure_telemetry, telemetry
from .updater import (
    check_for_updates,
    log_windows_update_message,
    self_upgrade,
)


def restart_mozphab():
    """Restart `moz-phab`, re-using the previous command line."""
    logger.info("Restarting...")

    # Explicitly close the log files to avoid issues with processes holding
    # exclusive logs on the files on Windows.
    stop_logging()

    # It's best to ignore errors here as they will be reported by the
    # new moz-phab process.
    p = subprocess.run(sys.argv)
    sys.exit(p.returncode)


def assert_api_token_is_present(repo: Repository, args: argparse.Namespace):
    """Assert a local API token is present.

    Prompt for install by running `install-certificate` if missing,
    unless the user is already running `install-certificate` themselves.
    """
    if args.command == "install-certificate":
        return

    try:
        conduit.load_api_token()
    except ConduitAPIError:
        logger.info("No API token detected, running `install-certificate`...")
        install = parse_args(["install-certificate"])
        install.func(repo, install)
        logger.info("Token installed, resuming original command")


def main(argv: List[str], *, is_development: bool):
    try:
        if not is_development and config.report_to_sentry:
            init_sentry()

        os.makedirs(environment.MOZBUILD_PATH, exist_ok=True)

        if config.no_ansi:
            environment.HAS_ANSI = False
        os.environ["MOZPHAB"] = "1"

        args = parse_args(argv)

        if args.trace:
            environment.DEBUG = True
        if environment.DEBUG:
            environment.SHOW_SPINNER = False

        init_logging()

        logger.debug("%s (%s)", environment.MOZPHAB_NAME, environment.MOZPHAB_VERSION)

        # Ensure that `patch --raw ..` only outputs the patch
        if args.command == "patch" and getattr(args, "raw", False):
            environment.SHOW_SPINNER = False
            logger.setLevel(logging.ERROR)

        elif args.command != "self-update":
            new_version = check_for_updates()

            if new_version and environment.IS_WINDOWS:
                log_windows_update_message()
            elif new_version and config.self_auto_update:
                try:
                    with wait_message(f"Upgrading to version {new_version}"):
                        self_upgrade()
                    restart_mozphab()
                except Exception as e:
                    logger.warning(
                        f"Failed to upgrade: {e};"
                        " continuing with current version..."
                        " Please report a bug if this behaviour persists."
                    )

        repo = None
        if args.needs_repo:
            with wait_message("Starting up.."):
                repo = repo_from_args(args)

            conduit.set_repo(repo)

        if not is_development:
            configure_telemetry(args)

        if repo is not None:
            assert_api_token_is_present(repo, args)

            telemetry().set_vcs(repo)
            try:
                args.func(repo, args)
            finally:
                repo.cleanup()

        else:
            args.func(args)

        telemetry().usage.command_time.stop()
        telemetry().submit()

    except KeyboardInterrupt:
        pass
    except ssl.SSLCertVerificationError as e:
        logger.error(e)

        if e.reason == "CERTIFICATE_VERIFY_FAILED":
            # Leave a helpful error message if the local issuer certificate
            # can't be found.
            logger.error(
                "\nYour Python installation is incomplete as you lack a local issuer "
                "certificate.\n"
                "See the following link for steps to resolve your problem:\n"
                "https://stackoverflow.com/questions/52805115/"
                "certificate-verify-failed-unable-to-get-local-issuer-certificate\n"
                "\n"
                "If you still experience trouble please file a bug."
            )

        sys.exit(1)
    except Error as e:
        logger.error(e)
        sys.exit(1)
    except Exception as e:
        if environment.DEBUG:
            logger.error(traceback.format_exc())
        else:
            logger.error("%s: %s", e.__class__.__name__, e)
            logger.error("Run moz-phab again with '--trace' to show debugging output")
        report_to_sentry(e)
        sys.exit(1)


def run():
    is_development = Version(environment.MOZPHAB_VERSION).is_prerelease
    main(sys.argv[1:], is_development=is_development)


if __name__ == "__main__":
    run()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import argparse
import os
from typing import Optional

from .exceptions import Error
from .git import Git
from .jujutsu import Jujutsu
from .mercurial import Mercurial
from .repository import Repository


def find_repo_root(path: str, avoid_jj: bool = False) -> Optional[str]:
    """Lightweight check for a repo in/under the specified path."""
    path = os.path.abspath(path)
    while os.path.split(path)[1]:
        if (
            Mercurial.is_repo(path)
            or (Jujutsu.is_repo(path) and not avoid_jj)
            or Git.is_repo(path)
        ):
            return path
        path = os.path.abspath(os.path.join(path, os.path.pardir))
    return None


def probe_repo(path: str, avoid_jj: bool = False) -> Optional[Repository]:
    """Attempt to find a repository at `path`."""
    try:
        return Mercurial(path)
    except ValueError:
        pass

    # NOTE: Jujutsu may have a co-located `.git` directory, so let's detect a `.jj` directory first.
    if not avoid_jj:
        try:
            return Jujutsu(path)
        except ValueError:
            pass

    try:
        return Git(path)
    except ValueError:
        pass

    return None


def repo_from_args(args: argparse.Namespace) -> Repository:
    """Returns a Repository object from either args.path or the cwd"""

    repo = None

    avoid_jj_vcs = hasattr(args, "avoid_jj_vcs") and args.avoid_jj_vcs

    def vcses_searched():
        jj = [] if avoid_jj_vcs else [".jj"]
        vcses_searched = [".hg"] + jj + [".git"]
        vcses_searched = " / ".join(vcses_searched)
        return vcses_searched

    # This allows users to override the below sanity checks.
    if hasattr(args, "path") and args.path:
        repo = probe_repo(args.path, avoid_jj=avoid_jj_vcs)
        if not repo:
            raise Error("%s: Not a repository: %s" % (args.path, vcses_searched()))

    else:
        # Walk parents to find repository root.
        path = find_repo_root(os.getcwd(), avoid_jj_vcs)
        if path:
            repo = probe_repo(path, avoid_jj_vcs)
        if not repo:
            raise Error(
                "Not a repository (or any of the parent directories): %s"
                % vcses_searched()
            )

    repo.set_args(args)
    return repo



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import argparse
import sys
from collections import OrderedDict
from typing import (
    Container,
    Dict,
    List,
    Optional,
    Tuple,
)

from mozphab.conduit import conduit
from mozphab.config import config
from mozphab.exceptions import Error
from mozphab.helpers import BUG_ID_RE, augment_commits_from_body, prompt
from mozphab.logger import logger
from mozphab.repository import Repository
from mozphab.spinner import wait_message
from mozphab.telemetry import telemetry


def create_hyperlink(text: str, url: str) -> str:
    """Create a terminal hyperlink using ANSI escape sequences."""
    return f"\033]8;;{url}\033\\{text}\033]8;;\033\\"


def linkify_revision_id(
    revision_id: str, phab_url: str, hyperlinks_enabled: bool = True
) -> str:
    """Convert revision ID like 'D123456' to a clickable Phabricator link."""
    if not hyperlinks_enabled:
        return revision_id
    # Extract the numeric part (remove 'D' prefix)
    numeric_id = revision_id.removeprefix("D")
    phabricator_url = f"{phab_url}/D{numeric_id}"
    return create_hyperlink(revision_id, phabricator_url)


def linkify_bugs_in_text(
    text: str, bmo_url: str, hyperlinks_enabled: bool = True
) -> str:
    """Find bug numbers in text and make them clickable Bugzilla links."""
    if not hyperlinks_enabled:
        return text

    def replace_bug(match):
        bug_number = match.group(1)
        bugzilla_url = f"{bmo_url}/show_bug.cgi?id={bug_number}"
        return create_hyperlink(f"Bug {bug_number}", bugzilla_url)

    return BUG_ID_RE.sub(replace_bug, text)


def to_llist(revisions: List[str]) -> Dict[str, Optional[str]]:
    """Converts a list to a linked list.

    Parameters:
        revisions (list): a list of revisions to create a linked list

    Returns (dict) a linked list in the form `dict(rev1="rev2", rev2="rev3", rev3=None)`
    where `rev2` is the child of `rev1` and the parent of `rev3` which has no children.
    """
    llist = {}
    for i, revision in enumerate(revisions):
        child = revisions[i + 1] if i + 1 < len(revisions) else None
        llist[revision] = child
    return llist


def walk_llist(
    llist: Dict[str, Optional[str]], allow_multiple_heads: bool = False
) -> List[str]:
    """Parse the llist for multiple heads and return a unique list of elements.

    Parameters:
        llist (dict): A linked list. {A_id: B_id, B_id: None}
        allow_multiple_heads (bool): Fail if multiple heads found.
    """
    referenced_children = [child for child in llist.values() if child]

    # Find head
    head = None
    for parent in sorted(llist.keys()):
        if parent not in referenced_children:
            if head:
                if not allow_multiple_heads:
                    raise Error("Multiple heads found.")
                break
            head = parent
    if not head:
        raise Error("Failed to find head.")

    # Walk list, checking for loops
    nodes = []
    while head:
        nodes.append(head)
        child = llist.get(head)
        if child and child in nodes:
            raise Error("Dependency loop")

        head = child

    return nodes


def remove_or_set_child(
    local_list: Dict, remote_list: Dict, revision: str
) -> Optional[Tuple[str, List[str]]]:
    """Return a `children.*` transaction for the revision based on remote/local state.

    Return `None` if no transaction is required.
    """
    # Set children if child is present.
    child = local_list[revision]
    if child is not None:
        return ("children.set", [child])

    # Remove child if the last child is a parent remotely
    if revision in remote_list and remote_list[revision] is not None:
        return ("children.remove", [remote_list[revision]])

    # No transaction otherwise.
    return None


def stack_transactions(
    remote_phids: List[str],
    local_phids: List[str],
    abandoned_revisions: Container[str],
    no_abandon: bool = False,
) -> Dict[str, List[Dict]]:
    """Prepare transactions to set the stack as provided in local_phids.

    Returns a dict of transactions for PHID as defined in
    `differential.revision.edit`.
    """
    remote_list = to_llist(remote_phids)
    local_list = to_llist(local_phids)

    transactions = OrderedDict()
    for revision in remote_phids + local_phids:
        transactions[revision] = []

    # Remote child no longer in stack
    remote_revisions_missing_from_local = [
        r for r in remote_phids if r not in local_phids
    ]
    for revision in remote_revisions_missing_from_local:
        child = remote_list[revision]
        if child is not None:
            transactions[revision].append(("children.remove", [child]))

        remote_list[revision] = None
        walk_llist(remote_list, allow_multiple_heads=True)

    # New revisions
    local_revisions_missing_from_remote = [
        r for r in local_phids if r not in remote_phids
    ]
    for revision in local_revisions_missing_from_remote:
        child = local_list[revision]

        transaction = remove_or_set_child(local_list, remote_list, revision)
        if transaction:
            transactions[revision].append(transaction)

        remote_list[revision] = child
        walk_llist(remote_list, allow_multiple_heads=True)

    # Modify child
    remote_revisions_found_in_local = [r for r in remote_phids if r in local_phids]
    for revision in remote_revisions_found_in_local:
        if local_list[revision] != remote_list[revision]:
            child = local_list[revision]

            transaction = remove_or_set_child(local_list, remote_list, revision)
            if transaction:
                transactions[revision].append(transaction)

            remote_list[revision] = child
            walk_llist(remote_list, allow_multiple_heads=True)

    # Abandon
    for revision in remote_revisions_missing_from_local:
        if no_abandon:
            # `--no-abandon` should cause revisions to never be abandoned.
            logger.debug(
                f"Skipping abandon transaction for {revision} due to `--no-abandon`."
            )
        elif revision not in abandoned_revisions:
            # Avoid abandoning a revision if it is already in `abandoned` state.
            transactions[revision].append(("abandon", True))

        del remote_list[revision]
        walk_llist(remote_list, allow_multiple_heads=True)

    assert ":".join(walk_llist(remote_list)) == ":".join(walk_llist(local_list))

    conduit_transactions = {}
    for revision, transaction_list in transactions.items():
        if not transaction_list:
            continue
        conduit_transactions.setdefault(revision, [])

        for trans_type, trans_value in transaction_list:
            conduit_transactions[revision].append(
                {"type": trans_type, "value": trans_value}
            )

    return conduit_transactions


def convert_stackgraph_to_linear(
    stack_graph: Dict[str, List[str]],
    phid_to_id: Dict[str, int],
) -> Dict[str, Optional[str]]:
    """Converts the `stackGraph` data from Phabricator to a linear format.

    Ensures each revision has only a single successor revision.
    """
    linear_stackgraph = {}

    for successor_phid, predecessor_phid_list in stack_graph.items():
        for predecessor_phid in predecessor_phid_list:
            # If we find a predecessor node in our linear stackgraph then
            # that node has multiple children.
            if predecessor_phid in linear_stackgraph:
                # Get the ID of the revision with multiple children.
                rev_id = phid_to_id[predecessor_phid]

                raise Error(f"Revision D{rev_id} has multiple children.")

            linear_stackgraph[predecessor_phid] = successor_phid

    # Set any heads as having no successor. Heads are in the values of
    # `linear_stackgraph` but not yet added as keys.
    heads = (
        successor
        # Use `list` here to avoid inspecting the list while we iterate over it,
        # causing a `RuntimeError: dictionary changed size during iteration`.
        for successor in list(linear_stackgraph.values())
        if successor not in linear_stackgraph
    )
    for head in heads:
        linear_stackgraph[head] = None

    return linear_stackgraph


def force_stack_transactions(
    remote_phids: List[str],
    local_phids: List[str],
    abandoned_revisions: Container[str],
    no_abandon_unconnected: bool = False,
) -> Dict[str, List[Dict]]:
    """Prepare transactions for force mode: synchronize remote to match local exactly.

    Force mode logic:
    1. Abandon all remote revisions not found locally in the range (unless no_abandon_unconnected is True)
    2. Unlink all revisions (remove all existing parent-children relationships)
    3. Relink revisions to match local state exactly

    Args:
        remote_phids: List of remote PHIDs in the stack
        local_phids: List of local PHIDs in the desired order
        abandoned_revisions: Set of already abandoned revision PHIDs
        no_abandon_unconnected: If True, do not abandon remote revisions not found locally

    Returns a dict of transactions for PHID as defined in `differential.revision.edit`.
    """
    local_list = to_llist(local_phids)
    transactions = OrderedDict()

    # Initialize transactions for all revisions that will be touched
    all_phids = set(remote_phids + local_phids)
    for revision in all_phids:
        transactions[revision] = []

    # Step 1: Abandon remote revisions not found locally
    remote_revisions_missing_from_local = [
        r for r in remote_phids if r not in local_phids
    ]

    if not no_abandon_unconnected:
        for revision in remote_revisions_missing_from_local:
            # Only abandon if not already abandoned
            if revision not in abandoned_revisions:
                transactions[revision].append(("abandon", True))

    # Step 2: Unlink all revisions - we'll use children.set with empty list to
    # remove all children. This effectively unlinks all existing relationships
    for revision in remote_phids:
        if revision not in local_phids and not no_abandon_unconnected:
            # Don't unset children for revisions that will be abandoned - no need to modify
            # their relationships since they'll be abandoned anyway
            continue
        transactions[revision].append(("children.set", []))

    # Step 3: Relink revisions to match local state exactly
    for revision in local_phids:
        child = local_list[revision]
        if child is not None:
            transactions[revision].append(("children.set", [child]))

    # Clean up empty transaction lists
    conduit_transactions = {}
    for revision, transaction_list in transactions.items():
        if not transaction_list:
            continue
        conduit_transactions.setdefault(revision, [])

        for trans_type, trans_value in transaction_list:
            conduit_transactions[revision].append(
                {"type": trans_type, "value": trans_value}
            )

    return conduit_transactions


def show_revision_glossary(
    transactions: Dict[str, List[Dict]],
    revisions: List[dict],
    repo: Repository,
    hyperlinks_enabled: bool = True,
):
    """Show a glossary with commit messages for all referenced revisions."""
    # Collect all referenced revision PHIDs
    referenced_phids = set()
    for phid, rev_transactions in transactions.items():
        referenced_phids.add(phid)
        for transaction in rev_transactions:
            if transaction["type"] in ["children.set", "children.remove"]:
                if transaction["value"]:
                    referenced_phids.update(transaction["value"])

    if not referenced_phids:
        return

    # Create a mapping from PHID to revision data
    phid_to_revision = {revision["phid"]: revision for revision in revisions}

    # Display the glossary
    logger.info("")
    logger.info("Referenced revisions:")
    for phid in sorted(
        referenced_phids,
        key=lambda p: phid_to_revision[p]["id"] if p in phid_to_revision else 0,
    ):
        if phid not in phid_to_revision:
            continue
        revision_id = f"D{phid_to_revision[phid]['id']}"
        linked_revision_id = linkify_revision_id(
            revision_id, repo.phab_url, hyperlinks_enabled
        )
        if phid in phid_to_revision:
            title = phid_to_revision[phid]["fields"].get("title", "(no title)")
            linked_title = linkify_bugs_in_text(title, repo.bmo_url, hyperlinks_enabled)
            logger.info(" * %s: %s", linked_revision_id, linked_title)
        else:
            logger.info(" * %s: (unknown)", linked_revision_id)


def format_stack(
    phids: List[str],
    title: str,
    phid_to_revision: Dict[str, dict],
    repo: Repository,
    hyperlinks_enabled: bool = True,
) -> None:
    """Format and display a stack of revisions."""
    if not phids:
        logger.info("%s: (empty)", title)
        return
    logger.info("%s:", title)
    # Reverse the order so bottom patch is at the bottom (stack order)
    reversed_phids = list(reversed(phids))
    for phid in reversed_phids:
        revision_id = f"D{phid_to_revision[phid]['id']}"
        linked_revision_id = linkify_revision_id(
            revision_id, repo.phab_url, hyperlinks_enabled
        )
        if phid in phid_to_revision:
            title_text = phid_to_revision[phid]["fields"].get("title", "(no title)")
            linked_title_text = linkify_bugs_in_text(
                title_text, repo.bmo_url, hyperlinks_enabled
            )
            logger.info("○ %s %s", linked_revision_id, linked_title_text)
        else:
            logger.info("○ %s (unknown)", linked_revision_id)


def show_verbose_stack_info(
    phabstack_phids: List[str],
    localstack_phids: List[str],
    revisions: List[dict],
    repo: Repository,
    hyperlinks_enabled: bool = True,
):
    """Show detailed information about the current and future stack shapes."""
    phid_to_revision = {revision["phid"]: revision for revision in revisions}

    logger.info("")
    logger.info("Stack reorganization details:")
    format_stack(
        phabstack_phids,
        "Current remote stack",
        phid_to_revision,
        repo,
        hyperlinks_enabled,
    )
    format_stack(
        localstack_phids,
        "Target local stack",
        phid_to_revision,
        repo,
        hyperlinks_enabled,
    )
    logger.info("")


def reorganise_inner(repo: Repository, args: argparse.Namespace):
    """Reorganise the stack on Phabricator to match the stack in the local VCS."""
    telemetry().submission.preparation_time.start()

    # Validate argument combinations
    if args.no_abandon_unconnected and not args.force:
        raise Error("--no-abandon-unconnected can only be used with --force")

    with wait_message("Checking connection to Phabricator."):
        # Check if raw Conduit API can be used
        if not conduit.check():
            raise Error("Failed to use Conduit API")

    # Find and preview commits to submits.
    with wait_message("Looking for commits.."):
        commits = repo.commit_stack()

    if not commits:
        raise Error("Failed to find any commits to reorganise.")

    with wait_message("Loading commits.."):
        augment_commits_from_body(commits)

    localstack_ids = [commit.rev_id for commit in commits]
    if not all(localstack_ids):
        if args.force:
            raise Error(
                "Force mode requires all local revisions to be present on Phabricator."
            )
        names = [commit.name for commit in commits if commit.rev_id is None]
        plural = len(names) > 1
        raise Error(
            "Found new commit{plural} in the local stack: {names}.\n"
            "Please submit {them} separately and call `moz-phab reorg` again.".format(
                plural="s" if plural else "",
                them="them" if plural else "it",
                names=", ".join(names),
            )
        )

    logger.warning(
        "Reorganisation based on {} commit{}:".format(
            len(commits),
            "" if len(commits) == 1 else "s",
        )
    )

    # Preload the phabricator stack
    with wait_message("Preloading Phabricator stack revisions..."):
        revisions = conduit.get_revisions(ids=localstack_ids)

    if not revisions:
        raise Error("Could not find revisions on Phabricator.")

    # Merge all the existing stackgraphs into one. Any repeated keys
    # will have the same values.
    stack_graph = {
        predecessor: successors
        for revision in revisions
        for predecessor, successors in revision["fields"]["stackGraph"].items()
    }

    # Fetch data about any revisions that are missing from local stack.
    revisions = conduit.get_revisions(phids=list(stack_graph.keys()))

    phid_to_id = {revision["phid"]: revision["id"] for revision in revisions}

    try:
        # Validate the `stackGraph` field from our remote revisions.
        phabstack = convert_stackgraph_to_linear(stack_graph, phid_to_id)
    except Error:
        logger.error("Remote stack is not linear.")
        raise

    if phabstack:
        try:
            phabstack_phids = walk_llist(phabstack)
        except Error:
            if args.force:
                # In force mode, we ignore remote stack structure issues
                logger.warning(
                    "Remote stack is not linear, but continuing in force mode.\n"
                    "Detected stack:\n{}".format(
                        " <- ".join(conduit.phids_to_ids(list(phabstack.keys())))
                    )
                )
                phabstack_phids = list(phabstack.keys())
            else:
                logger.error(
                    "Remote stack is not linear.\nDetected stack:\n{}".format(
                        " <- ".join(conduit.phids_to_ids(list(phabstack.keys())))
                    )
                )
                raise
    else:
        phabstack_phids = []

    localstack_phids = conduit.ids_to_phids(localstack_ids)
    abandoned_revisions = {
        revision["phid"]
        for revision in revisions
        if revision["fields"]["status"]["value"] == "abandoned"
    }

    if args.force:
        transactions = force_stack_transactions(
            phabstack_phids,
            localstack_phids,
            abandoned_revisions,
            no_abandon_unconnected=args.no_abandon_unconnected,
        )
    else:
        try:
            transactions = stack_transactions(
                phabstack_phids,
                localstack_phids,
                abandoned_revisions,
                no_abandon=args.no_abandon,
            )
        except Error:
            logger.error("Unable to prepare stack transactions.")
            raise

    if not transactions:
        logger.info("Reorganisation is not needed.")
        return

    # Determine hyperlinks setting from config and args
    hyperlinks_enabled = config.hyperlinks and not args.no_hyperlinks

    if args.force:
        logger.warning("Stack will be forcibly synchronized:")
    else:
        logger.warning("Stack will be reorganised:")

    for phid, rev_transactions in transactions.items():
        if phid not in phid_to_id:
            # Skip PHIDs that are not in our mapping (can happen in tests)
            continue
        node_id = f"D{phid_to_id[phid]}"
        linked_node_id = linkify_revision_id(node_id, repo.phab_url, hyperlinks_enabled)
        if any(transaction["type"] == "abandon" for transaction in rev_transactions):
            logger.info(" * {} will be abandoned".format(linked_node_id))
        else:
            for t in rev_transactions:
                if t["type"] == "children.set":
                    if t["value"] and t["value"][0] in phid_to_id:  # Has children
                        child_id = f"D{phid_to_id[t['value'][0]]}"
                        linked_child_id = linkify_revision_id(
                            child_id, repo.phab_url, hyperlinks_enabled
                        )
                        logger.info(
                            " * {child} will depend on {parent}".format(
                                child=linked_child_id,
                                parent=linked_node_id,
                            )
                        )
                    elif args.force:
                        logger.info(
                            " * {} will have all dependencies removed".format(
                                linked_node_id
                            )
                        )
                if t["type"] == "children.remove":
                    if t["value"] and t["value"][0] in phid_to_id:
                        child_id = f"D{phid_to_id[t['value'][0]]}"
                        linked_child_id = linkify_revision_id(
                            child_id, repo.phab_url, hyperlinks_enabled
                        )
                        logger.info(
                            " * {child} will no longer depend on {parent}".format(
                                child=linked_child_id,
                                parent=linked_node_id,
                            )
                        )

    # Show glossary with commit messages for all referenced revisions (but not in verbose mode)
    if not args.verbose:
        show_revision_glossary(transactions, revisions, repo, hyperlinks_enabled)
    else:
        # Show verbose information at the bottom
        show_verbose_stack_info(
            phabstack_phids, localstack_phids, revisions, repo, hyperlinks_enabled
        )

    telemetry().submission.preparation_time.stop()

    if args.yes:
        pass
    else:
        if args.force:
            res = prompt("Perform force synchronization", ["Yes", "No"])
        else:
            res = prompt("Perform reorganisation", ["Yes", "No"])
        if res == "No":
            sys.exit(1)

    telemetry().submission.process_time.start()

    with wait_message("Applying transactions..."):
        for phid, rev_transactions in transactions.items():
            conduit.apply_transactions_to_revision(
                rev_id=phid,
                transactions=rev_transactions,
            )

    telemetry().submission.process_time.stop()
    logger.info("Stack has been reorganised.")


def reorganise(repo: Repository, args: argparse.Namespace):
    """Reorganise the stack on Phabricator to match the stack in the local VCS."""
    try:
        reorganise_inner(repo, args)
    except Error as e:
        # Check if the `force` flag was not set and print a hint about --force mode
        if not args.force:
            logger.warning(
                "Reorganisation failed. You might try using --force to bypass "
                "stack structure checks and force synchronization."
            )
        raise e


def add_parser(parser):
    reorg_parser = parser.add_parser("reorg", help="Reorganise commits in Phabricator.")
    reorg_parser.add_argument(
        "--yes",
        "-y",
        action="store_true",
        help="Reorganise without confirmation (default: False).",
    )
    reorg_parser.add_argument(
        "--safe-mode",
        dest="safe_mode",
        action="store_true",
        help="Run VCS with only necessary extensions.",
    )
    reorg_parser.add_argument(
        "--upstream",
        "--remote",
        "-u",
        action="append",
        help=(
            "Git only: Set remote to detect the starting commit. Overrides "
            "`git.remote` in config."
        ),
    )
    reorg_parser.add_argument(
        "start_rev",
        nargs="?",
        default="(auto)",
        help="Start revision of range to reorganise (default: detected).",
    )
    reorg_parser.add_argument(
        "end_rev",
        nargs="?",
        default=".",
        help="End revision of range to reorganise (default: current commit).",
    )
    reorg_parser.add_argument(
        "--no-abandon",
        action="store_true",
        dest="no_abandon",
        help="Do not abandon revisions during reorg.",
    )
    reorg_parser.add_argument(
        "--force",
        action="store_true",
        help="Force synchronization: abandon remote revisions not found locally, "
        "unlink all revisions, then relink to match local state.",
    )
    reorg_parser.add_argument(
        "--no-abandon-unconnected",
        action="store_true",
        help="When used with --force, do not abandon remote revisions that are not "
        "connected to the local stack. Useful for managing multiple patch series.",
    )
    reorg_parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="Show detailed information about stack reorganization, including "
        "previous and future stack shapes.",
    )
    reorg_parser.add_argument(
        "--no-hyperlinks",
        action="store_true",
        help="Disable terminal hyperlinks for revision IDs and bug numbers.",
    )
    reorg_parser.set_defaults(func=reorganise, needs_repo=True)



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import json
import os
import sys
import time
import urllib.request
from pathlib import Path
from typing import Optional

from packaging.version import Version
from packaging.version import parse as parse_version
from setuptools import Distribution

from mozphab import environment

from .config import config
from .environment import MOZPHAB_VERSION
from .exceptions import Error
from .logger import logger
from .subprocess_wrapper import check_call

# Update every three days.
SELF_UPDATE_FREQUENCY_SECONDS = 24 * 3 * 60 * 60


def get_pypi_json() -> dict:
    """Get data about `MozPhab` from the JSON API endpoint."""
    url = "https://pypi.org/pypi/MozPhab/json"
    output = urllib.request.urlopen(urllib.request.Request(url), timeout=30).read()
    response = json.loads(output.decode("utf-8"))
    return response


def get_simple_json() -> dict:
    """Get data about `MozPhab` from the `simple` API endpoint."""
    url = "https://pypi.org/simple/MozPhab"
    request = urllib.request.Request(
        url, headers={"Accept": "application/vnd.pypi.simple.v1+json"}
    )
    output = urllib.request.urlopen(request, timeout=30).read()
    return json.loads(output.decode("utf-8"))


def parse_latest_prerelease_version(simple_json: dict) -> str:
    """Parse PyPI's API response for `moz-phab` to determine the latest version."""
    # Get all the returned `.tar.gz` file entries.
    filenames = (entry["filename"].lower() for entry in simple_json["files"])

    # The format is `MozPhab-<version>.tar.gz`, so remove the prefix and
    # suffix to get the version strings.
    versions = (
        filename[len("mozphab-") :][: -len(".tar.gz")]
        for filename in filenames
        if filename.startswith("mozphab-") and filename.endswith(".tar.gz")
    )

    # Turn each version string into a `Version`, get the max, then return as `str`.
    return str(max(Version(version) for version in versions))


def log_windows_update_message():
    """Log a message telling users to update `moz-phab` to the latest version."""
    logger.warning(
        "Please run `pip install --upgrade MozPhab` to ugprade to the latest version."
    )


def should_self_update(self_last_check: int, current_time: int) -> bool:
    """Return `True` if a self-update should be performed."""
    # Setting `updater.self_last_check` to a negative value disables self-update.
    if self_last_check < 0:
        logger.debug("Self-update is disabled, skipping check for new versions.")
        return False

    # Avoid self-update if our last check was before the frequency interval.
    time_since_last_check = current_time - self_last_check
    if time_since_last_check <= SELF_UPDATE_FREQUENCY_SECONDS:
        logger.debug(
            f"Time since last check is {round(time_since_last_check)} seconds, "
            "skipping check for new versions."
        )
        return False

    return True


def check_for_updates(force_check: bool = False) -> Optional[str]:
    """Check if an update is available for `moz-phab`.

    Log a message about the new version, return the version as a `str` if it is
    found or return `None`. Use `force_check` to check for updates even when the
    usual conditions aren't met.
    """
    logger.debug(f"Forced update check: {force_check}.")

    # Return if our check conditions aren't met.
    current_time_s = int(time.time())
    if not force_check and not should_self_update(
        config.self_last_check, current_time_s
    ):
        logger.debug("Not checking for updates.")
        return

    config.self_last_check = current_time_s
    current_version = MOZPHAB_VERSION
    pypi_json = get_pypi_json()
    pypi_info = pypi_json["info"]

    if not config.get_pre_releases:
        # Use the latest full release.
        pypi_version = pypi_info["version"]
    else:
        # Find the latest pre-release version manually since the "version" key
        # only contains the latest full release on PyPI.
        simple_json = get_simple_json()
        pypi_version = parse_latest_prerelease_version(simple_json)

    logger.debug(f"Versions - local: {current_version}, PyPI: {pypi_version}")

    # convert ">=3.6" to (3, 6)
    try:
        required_python_version = tuple(
            [int(i) for i in pypi_info["requires_python"][2:].split(".")]
        )
    except ValueError:
        required_python_version = ()

    if sys.version_info < required_python_version:
        raise Error(
            "Unable to upgrade to version {}.\n"
            "MozPhab requires Python in version {}".format(
                pypi_version, pypi_info["requires_python"]
            )
        )

    config.write()

    if parse_version(current_version) >= parse_version(pypi_version):
        logger.debug("update check not required")
        return

    logger.warning(f"Version {pypi_version} of `moz-phab` is now available")

    return pypi_version


def self_upgrade():
    """Upgrade ourselves with pip."""

    # Run pip using the current python executable to accommodate for virtualenvs
    command = (
        [sys.executable]
        + ["-m", "pip"]
        + ["install", "MozPhab"]
        + ["--upgrade"]
        + ["--no-cache-dir"]
        + ["--disable-pip-version-check"]
    )

    if config.get_pre_releases:
        command += ["--pre"]

    if not environment.DEBUG:
        command += ["--quiet"]

    # `sys.argv[0]` is the path to the script that was used to start python.
    # For example: `/home/connor/.virtualenvs/moz-phab-dev/bin/moz-phab`. Run
    # `.resolve()` to make sure we have a full path, and then the `parent` is
    # the directory for the script.
    script_dir = Path(sys.argv[0]).resolve().parent

    # If moz-phab was installed with --user, we need to pass it to pip
    # Create "install" setuptools command with --user to find the scripts_path
    d = Distribution()
    d.parse_config_files()
    i = d.get_command_obj("install", create=True)
    # Forcing the environment detected by Distribution to the --user one
    i.user = True
    i.prefix = i.exec_prefix = i.home = i.install_base = i.install_platbase = None
    i.finalize_options()
    # Checking if the moz-phab script is installed in user's scripts directory
    user_dir = Path(i.install_scripts).resolve()

    # Prevent self-update from failing when installing over OS-managed
    # installs; see bug 1876182.
    command_env = os.environ.copy()
    command_env["PIP_BREAK_SYSTEM_PACKAGES"] = "1"

    if script_dir == user_dir:
        command.append("--user")

    if environment.IS_WINDOWS:
        # Windows does not allow to remove the exe file of the running process.
        # Renaming the `moz-phab.exe` file to allow pip to install a new version.
        temp_exe = script_dir / "moz-phab-temp.exe"
        try:
            temp_exe.unlink()
        except FileNotFoundError:
            pass

        exe = script_dir / "moz-phab.exe"
        exe.rename(temp_exe)

        try:
            check_call(command)
        except Exception:
            temp_exe.rename(exe)
            raise

        if not exe.is_file():
            # moz-phab.exe is not created - install wasn't needed.
            temp_exe.rename(exe)

    else:
        check_call(command, env=command_env)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import json
import os
import urllib.parse
from typing import (
    List,
    Optional,
)

from mozphab import environment

from .commits import Commit
from .conduit import conduit
from .diff import Diff
from .exceptions import Error
from .helpers import (
    get_arcrc_path,
    read_json_field,
)
from .logger import logger
from .spinner import wait_message

MOZILLA_DOMAINS = {
    ".mozilla.org",
    ".mozilla.com",
    ".allizom.org",
}

# Mapping of known Phabricator URLs to Lando URLs.
LANDO_URL_MAPPING = {
    "https://phabricator.services.mozilla.com": "https://lando.moz.tools",
    "https://phabricator-dev.allizom.org": "https://dev.lando.nonprod.webservices.mozgcp.net",
    "https://phabricator.allizom.org": "https://stage.lando.nonprod.webservices.mozgcp.net",
}


def get_lando_url_for_phabricator(phab_url: str) -> str:
    """Return the Lando URL for the given Phabricator URL."""
    return LANDO_URL_MAPPING[phab_url.rstrip("/")]


def is_mozilla_phabricator(url: str) -> bool:
    """Return `True` if the `url` is a Mozilla owned domain."""
    phab_host = urllib.parse.urlparse(url).hostname
    if not phab_host:
        return False
    return any(phab_host.endswith(domain) for domain in MOZILLA_DOMAINS)


class Repository(object):
    def __init__(self, path: str, dot_path: str, phab_url: Optional[str] = None):
        self._phid = None
        self._phab_repo = None
        self._phab_vcs = None
        self.vcs = None
        self.path = path  # base repository directory
        self.dot_path = dot_path  # .hg/.git directory
        self._arcconfig_files = [
            os.path.join(self.dot_path, ".arcconfig"),
            os.path.join(self.path, ".arcconfig"),
        ]
        self.args = None
        self.phab_url = (phab_url or self._phab_url()).rstrip("/")
        self.api_url = self._api_url()
        self.call_sign = self._get_setting("repository.callsign")
        self.bmo_url = self._get_setting("bmo_url")

        if self.bmo_url:
            if not (
                urllib.parse.urlparse(self.bmo_url).scheme == "https"
                or environment.HTTP_ALLOWED
            ):
                raise Error("Only https connections are allowed.")
        elif is_mozilla_phabricator(self.phab_url):
            self.bmo_url = "https://bugzilla.mozilla.org"

    def is_worktree_clean(self) -> bool:
        """Check if the working tree is clean."""

    def before_submit(self):
        """Executed before the submit commit."""

    def after_submit(self):
        """Executed after the submit commit."""

    def _get_setting(self, key):
        """Read settings from .arcconfig"""
        value = read_json_field(self._arcconfig_files, [key])
        return value

    def _phab_url(self) -> str:
        """Determine the phab/conduit URL."""

        # In order of priority as per arc
        # FIXME: This should also check {.hg|.git}/arc/config, which is where
        # `arc set-config --local` writes to.  See bug 1497786.
        defaults_files = [get_arcrc_path()]
        if environment.IS_WINDOWS:
            defaults_files.append(
                os.path.join(
                    os.getenv("ProgramData", ""), "Phabricator", "Arcanist", "config"
                )
            )
        else:
            defaults_files.append("/etc/arcconfig")

        phab_url = (
            self._get_setting("phabricator.uri")
            or self._get_setting("conduit_uri")
            or read_json_field(defaults_files, ["config", "default"])
        )

        if not phab_url:
            raise Error("Failed to determine Phabricator URL (missing .arcconfig?)")
        return phab_url

    def cleanup(self):
        """Perform any repo-related cleanup tasks.

        May be called multiple times.
        If an exception is raised this is NOT called (to avoid dataloss)."""

    def finalize(self, commits):
        """Update the history after node changed."""

    def set_args(self, args):
        if (
            hasattr(args, "single")
            and args.single
            and args.end_rev != environment.DEFAULT_END_REV
        ):
            raise Error("Option --single can be used with only one identifier.")

        self.args = args

    def untracked(self) -> List[str]:
        """Return a list of untracked files."""

    def commit_stack(self, **kwargs) -> Optional[List[Commit]]:
        """Return list of commits.

        List of `Commit`s:
            name          human readable identifier of commit (eg. short sha)
            node          SHA1 in stack
            orig-node     an original SHA1 of the commit
            title         first line of commit description (unaltered)
            body          commit description, excluding first line
            title-preview title with bug-id and reviewer modifications
            bug-id        bmo bug-id
            bug-id-orig   original bug-id from commit desc
            reviewers     list of reviewers
            rev-id        phabricator revision id
            parent        SHA1 of the parent commit
            author-date   string representation of the commit creation time
            author-date-epoch
            author-name
            author-email
        """

    def get_diff(self, commit: Commit) -> Diff:
        """Create a Diff object with changes."""

    def refresh_commit_stack(self, commits):
        """Update the stack following an altering change (eg rebase)."""

    def is_node(self, node):
        """Check if node exists.

        Returns a Boolean.
        """

    def check_node(self, node):
        """Check if node exists.

        Returns a node if found.

        Raises NotFoundError if node not found in the repository.
        """

    def checkout(self, node):
        """Checkout/Update to specified node."""

    def commit(self, body):
        """Commit the changes in the working directory."""

    def amend_commit(self, commit, commits):
        """Amend commit description from `title` and `desc` fields"""

    def is_descendant(self, node: str) -> bool:
        """Return `True` if the repository revset is descendant from `node`."""

    def get_repo_head_branch(self) -> Optional[str]:
        """Return the expected branch/head for the current Phabricator repo.

        Confirms the identified head exists in the repository.
        """

    def uplift_commits(self, dest: str, commits: List[Commit]) -> List[Commit]:
        """Uplift the repo's revset onto `dest` and returns the refreshed `commits`."""

    def rebase_commit(self, source_commit, dest_commit):
        """Rebase source onto destination."""

    def before_patch(self, node, name):
        """Prepare repository to apply the patches."""

    def apply_patch(
        self, diff: str, body: str, author: Optional[str], author_date: Optional[int]
    ):
        """Apply the patch and commit the changes."""

    def format_patch(
        self, diff: str, body: str, author: Optional[str], author_date: Optional[int]
    ):
        """Format a patch appropriate for importing."""

    def check_commits_for_submit(self, commits: List[Commit]):
        """Validate the list of commits are okay to submit."""

    def _api_url(self):
        """Return a base URL for conduit API call"""
        url = urllib.parse.urljoin(self.phab_url, "api/")

        if not (
            urllib.parse.urlparse(url).scheme == "https" or environment.HTTP_ALLOWED
        ):
            raise Error("Only https connections are allowed.")

        return url

    @property
    def phab_repo(self):
        """Representation of the Repository in Phabricator API."""
        if not self._phab_repo:
            with wait_message("Reading repository data"):
                self._phab_repo = conduit.get_repository_by_callsign(self.call_sign)

        return self._phab_repo

    @property
    def phid(self):
        """PHID of the repository.

        This value does not change over time.
        It is stored in a file to avoid calling the API on every run.
        """
        if not self._phid:
            path = os.path.join(self.dot_path, ".moz-phab_phid_cache")

            if os.path.isfile(path):
                with open(path) as f:
                    try:
                        repo_phids = json.load(f)
                    except json.decoder.JSONDecodeError:
                        # File is probably using the old format.
                        repo_phids = {}
                    repo_phid = repo_phids.get(self.call_sign, None)
            else:
                repo_phids = {}
                repo_phid = None

            if not repo_phid:
                repo_phid = self.phab_repo["phid"]
                repo_phids[self.call_sign] = repo_phid
                with open(path, "w") as f:
                    json.dump(repo_phids, f)
            self._phid = repo_phid

        return self._phid

    def check_vcs(self):
        """`Git.check_vcs` raises if cinnabar required and not installed."""
        if self.args.force_vcs:
            return True

        if self.vcs != self.phab_vcs:
            # This error is captured in Git and not raised if Cinnabar installed.
            logger.warning(
                "Local VCS ({local}) is different from the one defined in the "
                "repository ({remote}).".format(local=self.vcs, remote=self.phab_vcs)
            )

        return True

    @property
    def phab_vcs(self):
        """Version Control System short name stored in Phabricator.

        This value does not change over time.
        It is stored in a file to avoid calling the API on every run.
        """
        if not self._phab_vcs:
            # check file
            path = os.path.join(self.dot_path, ".moz-phab_vcs_cache")
            if os.path.isfile(path):
                with open(path) as f:
                    self._phab_vcs = f.readline()
            else:
                self._phab_vcs = self.phab_repo["fields"]["vcs"]
                with open(path, "w") as f:
                    f.write(self._phab_vcs)

        return self._phab_vcs

    @property
    def lando_url(self) -> str:
        """Return the Lando URL for this repository."""
        return get_lando_url_for_phabricator(self.phab_url)

    def get_public_node(self, node):
        """Hashtag in a remote VCS."""
        return node

    def validate_email(self):
        """Validate a user's configured email address."""



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import copy
from unittest import mock

import pytest
from packaging.version import Version

from mozphab import diff, environment, exceptions, mozphab
from mozphab.commits import Commit
from mozphab.diff import Diff
from mozphab.mercurial import Mercurial

from .conftest import assert_attributes, create_temp_fn


@mock.patch("mozphab.mercurial.Mercurial.hg_out")
def test_get_successor(m_hg_hg_out, hg):
    m_hg_hg_out.return_value = []
    assert (None, None) == hg._get_successor("x")

    m_hg_hg_out.return_value = ["1 abcde"]
    assert ["1", "abcde"] == hg._get_successor("x")

    m_hg_hg_out.return_value = ["a", "b"]
    with pytest.raises(exceptions.Error):
        hg._get_successor("x")


@mock.patch("mozphab.mercurial.Mercurial._get_successor")
@mock.patch("mozphab.mercurial.Mercurial.rebase_commit")
@mock.patch("mozphab.mercurial.Mercurial._get_parent")
def test_finalize(m_get_parent, m_hg_rebase, m_hg_get_successor, hg):
    commits = [
        Commit(node="aaa", orig_node="aaa"),
        Commit(node="bbb", orig_node="bbb"),
        Commit(node="ccc", orig_node="ccc"),
    ]

    m_get_parent.return_value = "different:than_others"
    m_hg_get_successor.return_value = (None, None)
    hg.finalize(copy.deepcopy(commits))
    assert m_hg_rebase.call_count == 2
    assert m_hg_rebase.call_args_list == [
        mock.call(
            Commit(node="bbb", orig_node="bbb"),
            Commit(node="aaa", orig_node="aaa"),
        ),
        mock.call(
            Commit(node="ccc", orig_node="ccc"),
            Commit(node="bbb", orig_node="bbb"),
        ),
    ]

    m_get_parent.side_effect = ("first", "aaa", "last")
    m_hg_rebase.reset_mock()
    hg.finalize(commits)
    m_hg_rebase.assert_called_once_with(
        Commit(node="ccc", orig_node="ccc"),
        Commit(node="bbb", orig_node="bbb"),
    )

    m_hg_get_successor.reset_mock()
    m_get_parent.side_effect = None
    m_get_parent.return_value = "different:than_others"
    m_hg_get_successor.side_effect = [(None, None), ("4", "ddd")]
    _commits = commits[:]
    hg.finalize(_commits)
    assert m_hg_get_successor.call_count == 2
    assert m_hg_get_successor.call_args_list == [mock.call("bbb"), mock.call("ccc")]
    assert _commits == [
        Commit(node="aaa", orig_node="aaa"),
        Commit(node="bbb", orig_node="bbb"),
        Commit(node="ddd", orig_node="ccc", name="4:ddd"),
    ]

    m_hg_rebase.reset_mock()
    m_hg_get_successor.side_effect = None
    m_hg_get_successor.return_value = (None, None)
    _commits = commits[:]
    _commits[0].node = "AAA"  # node has been amended
    hg.finalize(_commits)
    assert m_hg_rebase.call_count == 2


@mock.patch("mozphab.mercurial.Mercurial.rebase_commit")
def test_finalize_no_evolve(m_hg_rebase, hg):
    hg.use_evolve = False
    hg.finalize([{"rev": "1", "node": "aaa"}, {"rev": "2", "node": "bbb"}])
    m_hg_rebase.assert_not_called()


@mock.patch("mozphab.mercurial.parse_config")
@mock.patch("mozphab.mercurial.Mercurial.hg_out")
@mock.patch("mozphab.mercurial.Mercurial.hg_log")
@mock.patch("mozphab.mercurial.hglib.open")
def test_set_args(m_hglib_open, m_hg_hg_log, m_hg_hg_out, m_parse_config, hg):
    class Args:
        def __init__(self, start="(auto)", end=".", safe_mode=False, single=False):
            self.start_rev = start
            self.end_rev = end
            self.safe_mode = safe_mode
            self.single = single

    m_config = mozphab.config
    with pytest.raises(exceptions.Error):
        hg.set_args(Args())

    # baseline config
    hg.mercurial_version = Version("4.5")
    m_config.safe_mode = False
    m_parse_config.return_value = {
        "ui.username": "username",
        "extensions.evolve": "",
        "extensions.topic": "",
    }

    # evolve & topic & shelve
    hg._hg = []
    hg.set_args(Args())

    assert "extensions.rebase" in hg._config_options
    assert hg._config_options["rebase.experimental.inmemory"] == "true"
    assert hg._extra_options["--pager"] == "never"
    assert hg.use_evolve
    assert hg.use_topic
    assert not hg.has_shelve

    # inmemory rebase requires hg 4.5+
    hg.mercurial_version = Version("4.0")
    hg._hg = []
    hg.set_args(Args())
    assert "extensions.rebase" in hg._config_options
    assert "rebase.experimental.inmemory" not in hg._config_options
    assert hg._extra_options["--pager"] == "never"
    hg.mercurial_version = Version("4.5")

    # safe_mode
    hg._hg = []
    hg.set_args(Args(safe_mode=True))
    options = hg._get_config_options()
    assert options == [
        ("extensions.rebase", ""),
        ("ui.username", "username"),
        ("extensions.evolve", ""),
        ("extensions.topic", ""),
    ]

    m_config.safe_mode = True
    hg._hg = []
    hg.set_args(Args())
    options = hg._get_config_options()
    assert options == [
        ("extensions.rebase", ""),
        ("ui.username", "username"),
        ("extensions.evolve", ""),
        ("extensions.topic", ""),
    ]
    m_config.safe_mode = False

    # no evolve
    m_parse_config.return_value = {"ui.username": "username", "extensions.shelve": ""}
    hg._hg = []
    hg.set_args(Args())
    options = hg._get_config_options()
    assert options == [
        ("extensions.rebase", ""),
        ("rebase.experimental.inmemory", "true"),
        ("experimental.evolution.createmarkers", "true"),
        ("extensions.strip", ""),
    ]
    assert not hg.use_evolve
    assert hg.has_shelve

    m_hg_hg_log.side_effect = [("123456789012",), ("098765432109",)]
    hg._hg = []
    hg.set_args(Args())
    assert "123456789012::098765432109" == hg.revset

    m_hg_hg_log.side_effect = IndexError
    with pytest.raises(exceptions.Error):
        hg.set_args(Args())

    m_hg_hg_log.reset_mock()
    m_hg_hg_log.side_effect = [("123456789012",), ("123456789012",)]
    hg.set_args(Args(single=True))
    assert "123456789012" == hg.revset
    assert m_hg_hg_log.call_args_list == [mock.call(".")]

    m_hg_hg_log.reset_mock()
    m_hg_hg_log.side_effect = [("123456789012",), ("123456789012",)]
    hg.set_args(Args(start="start", single=True))
    assert "123456789012" == hg.revset
    assert m_hg_hg_log.call_args_list == [mock.call("start")]


@mock.patch("mozphab.mercurial.Mercurial._status")
def test_clean_worktree(m_status, hg):
    m_status.return_value = {"T": None, "U": None}
    assert hg.is_worktree_clean()

    m_status.return_value = {"T": True, "U": None}
    assert not hg.is_worktree_clean()

    m_status.return_value = {"T": None, "U": True}
    assert hg.is_worktree_clean()

    m_status.return_value = {"T": True, "U": True}
    assert not hg.is_worktree_clean()


@mock.patch("mozphab.mercurial.Mercurial.hg")
def test_commit(m_hg, hg):
    hg.commit("some body")
    m_hg.assert_called_once()


@mock.patch("mozphab.mercurial.Mercurial.checkout")
@mock.patch("mozphab.mercurial.Mercurial.hg_out")
@mock.patch("mozphab.mercurial.Mercurial.hg")
@mock.patch("mozphab.mercurial.config")
def test_before_patch(m_config, m_hg, m_hg_out, m_checkout, hg):
    class Args:
        def __init__(
            self,
            rev_id="D123",
            nocommit=False,
            raw=False,
            applyto="base",
            no_bookmark=False,
            no_topic=False,
        ):
            self.rev_id = rev_id
            self.nocommit = nocommit
            self.raw = raw
            self.applyto = applyto
            self.no_bookmark = no_bookmark
            self.no_topic = no_topic

    m_config.create_bookmark = True
    m_config.create_topic = False
    m_hg_out.side_effect = ["bookmark"]
    hg.args = Args()
    hg.before_patch("sha1", "bookmark")
    m_checkout.assert_called_with("sha1")
    m_hg_out.assert_called()
    m_hg.assert_called_with(["bookmark", "bookmark_1"])
    m_checkout.assert_called_once_with("sha1")

    m_checkout.reset_mock()
    hg.args = Args(nocommit=True)
    m_hg.reset_mock()
    hg.before_patch("sha1", None)
    m_hg.assert_not_called()
    m_checkout.assert_called_once_with("sha1")

    hg.args = Args(applyto="here")
    m_checkout.reset_mock()
    m_hg_out.reset_mock()
    m_hg_out.side_effect = None
    m_hg_out.return_value = "some book_marks"
    hg.before_patch(None, "bookmark")
    m_hg_out.assert_called_once()
    m_checkout.assert_not_called()

    hg.args = Args(applyto="here")
    m_checkout.reset_mock()
    m_hg_out.reset_mock()
    m_hg_out.side_effect = None
    m_hg_out.return_value = "some book_marks"
    hg.before_patch(None, "bookmark")
    m_hg_out.assert_called_once()
    m_checkout.assert_not_called()

    m_hg_out.reset_mock()
    hg.args = Args(no_bookmark=True)
    hg.before_patch(None, "bookmark")
    m_hg_out.assert_not_called()

    m_config.create_bookmark = False
    m_hg_out.reset_mock()
    hg.args = Args()
    hg.before_patch(None, "bookmark")
    m_hg_out.assert_not_called()

    # Create a topic, but requested topic collides with an existing
    # topic and so must be renamed.
    m_config.create_topic = True
    hg.use_topic = True
    m_hg_out.reset_mock()
    m_hg_out.side_effect = ["topic_1"]
    hg.args = Args()
    hg.before_patch("sha1", "topic")
    m_checkout.assert_called_with("sha1")
    m_hg_out.assert_called()
    m_hg.assert_called_with(["topic", "topic_2"])
    m_checkout.assert_called_once_with("sha1")

    # No conflict => use requested name.
    m_hg_out.side_effect = ["topic"]
    hg.before_patch("sha1", "other")
    m_hg.assert_called_with(["topic", "other"])

    # Create both a bookmark and a topic. Not a useful thing to do,
    # but it should work. And bookmark names and topic names should not
    # collide when looking for pre-existing names.

    m_config.create_bookmark = True
    m_hg.reset_mock()  # Clear the calls.
    m_hg_out.side_effect = ["bookmark", "topic"]
    hg.before_patch("sha1", "bookmark")
    m_hg.assert_has_calls(
        [
            mock.call(["bookmark", "bookmark_1"]),
            mock.call(["topic", "bookmark"]),
        ]
    )


@mock.patch("mozphab.mercurial.temporary_binary_file")
@mock.patch("mozphab.mercurial.Mercurial.hg")
def test_apply_patch(m_hg, m_temp_bin_fn, hg):
    m_temp_bin_fn.return_value = create_temp_fn("diff_fn")
    hg.apply_patch("diff", "body", "user", 1)
    m_hg.assert_called_once_with(["import", "diff_fn", "--quiet"])
    m_temp_bin_fn.assert_called_once_with(
        b"# HG changeset patch\n# User user\n# Date 1 0\nbody\ndiff"
    )


@mock.patch("mozphab.mercurial.Mercurial.hg_out")
def test_is_node(m_hg_out, hg):
    assert hg.is_node("aabbcc")
    m_hg_out.assert_called_once_with(["identify", "-q", "-r", "aabbcc"])

    m_hg_out.side_effect = mock.Mock(side_effect=exceptions.CommandError)
    assert not hg.is_node("aaa")


@mock.patch("mozphab.mercurial.Mercurial.hg_out")
def test_is_descendant(m_hg_out, hg):
    m_hg_out.return_value = ""
    assert (
        hg.is_descendant("aabbcc") is False
    ), "Empty log result should indicated commit is not a descendant."

    m_hg_out.return_value = "aabbcc"
    assert (
        hg.is_descendant("aabbcc") is True
    ), "Non-empty log result should indicated commit is a descendant."


@mock.patch("mozphab.mercurial.Mercurial.is_node")
def test_check_node(m_is_node, hg):
    node = "aabbcc"
    m_is_node.return_value = True
    assert node == hg.check_node(node)

    m_is_node.return_value = False
    with pytest.raises(exceptions.NotFoundError) as e:
        hg.check_node(node)

    assert "" == str(e.value)


@mock.patch("mozphab.mercurial.Mercurial.hg_out")
def test_hg_cat(m_hg, hg):
    cat = m_hg.return_value = b"some text"
    hg.hg_cat("fn", "node")
    m_hg.assert_called_once_with(
        ["cat", "-r", "node", "fn"], expect_binary=True, split=False
    )
    assert cat == b"some text"


@mock.patch("mozphab.mercurial.Mercurial.hg_out")
def test_file_size(m_hg, hg):
    m_hg.return_value = "123\n"
    res = hg._file_size("fn", "rev")
    m_hg.assert_called_once_with(
        ["files", "-v", "-r", "rev", mock.ANY, "-T", "{size}"], split=False
    )
    assert res == 123


@mock.patch("mozphab.mercurial.Mercurial._file_size")
@mock.patch("mozphab.mercurial.Mercurial.hg_cat")
def test_file_meta(m_cat, m_file_size, hg):
    size = environment.MAX_TEXT_SIZE - 1
    m_file_size.return_value = size
    m_cat.return_value = b"spam\nham"
    meta = hg._get_file_meta("fn", "rev")
    assert meta == {
        "binary": False,
        "mime": "TEXT",
        "bin_body": b"spam\nham",
        "body": "spam\nham",
        "file_size": size,
    }


@mock.patch("mozphab.mercurial.mimetypes")
@mock.patch("mozphab.mercurial.Mercurial._file_size")
@mock.patch("mozphab.mercurial.Mercurial.hg_cat")
def test_file_meta_binary(m_cat, m_file_size, m_mime, hg):
    m_mime.guess_type.return_value = ["MIMETYPE"]
    m_cat.return_value = b"spam\nham"
    size = environment.MAX_TEXT_SIZE + 1
    m_file_size.return_value = size
    meta = hg._get_file_meta("fn", "rev")
    assert meta == {
        "binary": True,
        "mime": "MIMETYPE",
        "bin_body": b"spam\nham",
        "body": b"spam\nham",
        "file_size": size,
    }

    hg._get_file_meta.cache_clear()
    hg.hg_cat.cache_clear()
    hg._file_size.cache_clear()

    size = environment.MAX_TEXT_SIZE - 1
    m_file_size.return_value = size
    m_cat.return_value = b"\0spam\nham"
    meta = hg._get_file_meta("fn", "rev")
    assert meta == {
        "binary": True,
        "mime": "MIMETYPE",
        "bin_body": b"\0spam\nham",
        "body": b"\0spam\nham",
        "file_size": size,
    }


@mock.patch("mozphab.mercurial.Mercurial._get_file_meta")
@mock.patch("mozphab.diff.Diff.Change.set_as_binary")
def test_change_add(m_set_as_binary, m_get_file_meta, hg):
    change = Diff.Change("x")
    m_get_file_meta.return_value = {
        "binary": False,
        "bin_body": b"abc\n",
        "body": "abc\n",
        "file_size": 123,
    }
    hg._change_add(change, "fn", None, "parent", "node")
    m_set_as_binary.assert_not_called()
    assert len(change.hunks) == 1
    assert_attributes(
        change.hunks[0],
        {
            "corpus": "+abc\n",
            "old_off": 0,
            "new_off": 1,
            "old_len": 0,
            "new_len": 1,
        },
    )

    # binary
    change = Diff.Change("x")
    m_set_as_binary.reset_mock()
    m_get_file_meta.return_value = {
        "binary": True,
        "bin_body": b"abc\n",
        "body": "abc\n",
        "file_size": 123,
        "mime": "MIME",
    }
    hg._change_add(change, "fn", None, "parent", "node")
    assert not change.hunks
    m_set_as_binary.assert_called_once_with(
        a_body="",
        a_mime="",
        b_body=b"abc\n",
        b_mime="MIME",
    )

    # empty
    change = Diff.Change("x")
    m_set_as_binary.reset_mock()
    m_get_file_meta.return_value = {
        "binary": False,
        "bin_body": b"",
        "body": "",
        "file_size": 0,
    }
    hg._change_add(change, "fn", None, "parent", "node")
    assert not change.hunks
    m_set_as_binary.assert_not_called()


@mock.patch("mozphab.mercurial.Mercurial._get_file_meta")
@mock.patch("mozphab.diff.Diff.Change.set_as_binary")
def test_change_del(m_set_as_binary, m_get_file_meta, hg):
    change = Diff.Change("x")
    m_get_file_meta.return_value = {
        "binary": False,
        "bin_body": b"abc\n",
        "body": "abc\n",
        "file_size": 123,
    }
    hg._change_del(change, "fn", None, "parent", "node")
    assert len(change.hunks) == 1
    assert_attributes(
        change.hunks[0],
        {
            "corpus": "-abc\n",
            "old_off": 1,
            "new_off": 0,
            "old_len": 1,
            "new_len": 0,
        },
    )
    m_set_as_binary.assert_not_called()

    # binary
    change = Diff.Change("x")
    m_set_as_binary.reset_mock()
    m_get_file_meta.return_value = {
        "binary": True,
        "bin_body": b"abc\n",
        "body": "abc\n",
        "file_size": 123,
        "mime": "MIME",
    }
    hg._change_del(change, "fn", None, "parent", "node")
    assert not change.hunks
    m_set_as_binary.assert_called_once_with(
        a_body=b"abc\n",
        a_mime="MIME",
        b_body="",
        b_mime="",
    )

    # empty
    change = Diff.Change("x")
    m_set_as_binary.reset_mock()
    m_get_file_meta.return_value = {
        "binary": False,
        "bin_body": b"",
        "body": "",
        "file_size": 0,
    }
    hg._change_del(change, "fn", None, "parent", "node")
    assert not change.hunks
    m_set_as_binary.assert_not_called()


@mock.patch("mozphab.mercurial.Mercurial._get_file_meta")
@mock.patch("mozphab.diff.Diff.Change.set_as_binary")
@mock.patch("mozphab.diff.Diff.Change.from_git_diff")
@mock.patch("mozphab.mercurial.Mercurial.hg_out")
def test_change_mod(m_hg_out, m_from_git_diff, m_set_as_binary, m_get_file_meta, hg):
    class Args:
        def __init__(self, lesscontext=False):
            self.lesscontext = lesscontext

    # file contents changed
    change = diff.Diff.Change(None)
    text_side_effect = (
        {"binary": False, "bin_body": b"abc\n", "body": "abc\n", "file_size": 4},
        {"binary": False, "bin_body": b"def\n", "body": "def\n", "file_size": 4},
    )
    m_get_file_meta.side_effect = text_side_effect
    m_hg_out.return_value = b"""\
diff --git a/fn b/fn
--- a/B
+++ b/B
@@ -1,1 +1,1 @@
-abc
+def
"""
    hg.args = Args()
    hg._change_mod(change, "fn", "old_fn", "parent", "node")
    m_hg_out.assert_called_once_with(
        ["diff"]
        + ["--git"]
        + ["--unified", str(environment.MAX_CONTEXT_SIZE)]
        + ["--rev", "parent"]
        + ["fn"],
        expect_binary=True,
    )
    m_from_git_diff.assert_called_once()

    # --less-content honoured
    m_hg_out.reset_mock()
    m_get_file_meta.side_effect = text_side_effect
    hg.args = Args(lesscontext=True)
    hg._change_mod(change, "fn", "old_fn", "parent", "node")
    m_hg_out.assert_called_once_with(
        ["diff", "--git", "--unified", "100", "--rev", "parent", "fn"],
        expect_binary=True,
    )

    # binary
    m_from_git_diff.reset_mock()
    m_get_file_meta.side_effect = (
        {
            "binary": True,
            "bin_body": b"abc\n",
            "body": b"abc\n",
            "file_size": 4,
            "mime": "MIME",
        },
        {
            "binary": False,
            "bin_body": b"def\n",
            "body": "def\n",
            "file_size": 4,
            "mime": "TEXT",
        },
    )
    hg._change_mod(change, "fn", "old_fn", "parent", "node")
    m_set_as_binary.assert_called_once_with(
        a_body=b"abc\n", a_mime="MIME", b_body=b"def\n", b_mime="TEXT"
    )


@mock.patch("mozphab.mercurial.Mercurial.hg_out")
def test_get_file_modes(m_hg, hg):
    m_hg.side_effect = (
        ["  file name"],  # status
        [" :file name"],  # files - parent
        [" :file name"],  # files - node
    )
    actual = hg._get_file_modes(Commit(node="aaa", parent="bbb"))
    expected = {"file name": {"old_mode": "100644", "new_mode": "100644"}}
    assert actual == expected

    m_hg.side_effect = (
        ["M file name"],  # status
        [" :file name"],  # files - parent
        ["x:file name"],  # files - node
    )
    actual = hg._get_file_modes(Commit(node="aaa", parent="bbb"))
    expected = {"file name": {"old_mode": "100644", "new_mode": "100755"}}
    assert actual == expected

    m_hg.side_effect = (
        ["A file name"],  # status
        [],  # files - parent
        ["x:file name"],  # files - node
    )
    actual = hg._get_file_modes(Commit(node="aaa", parent="bbb"))
    expected = {"file name": {"new_mode": "100755"}}
    assert actual == expected

    m_hg.side_effect = (
        ["R file name"],  # status
        [" :file name"],  # files - parent
        [],  # files - node
    )
    actual = hg._get_file_modes(Commit(node="aaa", parent="bbb"))
    expected = {"file name": {"old_mode": "100644"}}
    assert actual == expected


def test_check_vcs(hg):
    class Args:
        def __init__(self, force_vcs=False):
            self.force_vcs = force_vcs

    hg.args = Args()
    assert hg.check_vcs()

    hg._phab_vcs = "git"

    # This just produces a warning now.
    # See bug 1963878.
    hg.check_vcs()

    hg.args = Args(force_vcs=True)
    assert hg.check_vcs()


@mock.patch("mozphab.mercurial.hglib.open")
@mock.patch("mozphab.repository.Repository._phab_url")
@mock.patch("mozphab.repository.os.chdir")
@mock.patch("os.path.isdir")
@mock.patch("mozphab.helpers.which")
def test_repository_cached(m_which, m_is_dir, m_os_chdir, m_phab_url, m_open, *patched):
    class MyRepo:
        version = 4, 7, 3

        def rawcommand(self, *args, **kw):
            return b"ui.username=xxx"

        def close(self):
            pass

    m_is_dir.return_value = True
    m_os_chdir.return_value = True
    m_phab_url.return_value = ""
    m_which.return_value = True
    m_open.return_value = MyRepo()

    hg = Mercurial("x")

    class Args:
        def __init__(self):
            self.start_rev = "(auto)"
            self.end_rev = "."
            self.safe_mode = True
            self.single = False

    hg._repo = None
    hg.set_args(Args())
    current_repo = hg.repository
    assert current_repo is not None

    # makes sure we cache the `hgclient` instance when using the
    # same args
    hg.set_args(Args())
    assert hg.repository is current_repo


@mock.patch("mozphab.mercurial.Mercurial.is_node")
def test_hg_get_repo_head_branch(m_is_node, hg):
    m_is_node.return_value = False
    assert (
        hg.get_repo_head_branch() is None
    ), "Unknown head should have returned `None`."

    m_is_node.return_value = True
    assert hg.get_repo_head_branch() == "beta", "beta did not correctly map to a branch"


def test_hg_validate_email(hg):
    with pytest.raises(exceptions.Error):
        hg.validate_email()

    hg.username = "Test User <test@mozilla.com>"
    assert hg.validate_email() is None, "validate_email() executes without error"


def test_hg_extract_email_from_username(hg):
    # Empty username
    hg.username = ""
    assert hg.extract_email_from_username() == ""

    # Username without email
    hg.username = "test user"
    assert hg.extract_email_from_username() == "test user"

    # Username with email
    hg.username = "test user <test@example.com>"
    assert hg.extract_email_from_username() == "test@example.com"
